from __future__ import annotations

import importlib.util
import os
import secrets
import shutil
import stat
import string
from contextlib import contextmanager, suppress
from pathlib import Path
from re import compile as regexp
from textwrap import indent
from typing import TYPE_CHECKING, Tuple

from git import Git, Repo
from pydantic.dataclasses import dataclass

from semantic_release.changelog.context import ChangelogMode, make_changelog_context
from semantic_release.changelog.release_history import ReleaseHistory
from semantic_release.commit_parser._base import CommitParser, ParserOptions
from semantic_release.commit_parser.conventional import ConventionalCommitParser
from semantic_release.commit_parser.token import (
    ParsedCommit,
    ParsedMessageResult,
    ParseError,
    ParseResult,
)
from semantic_release.enums import LevelBump

from tests.const import SUCCESS_EXIT_CODE

if TYPE_CHECKING:
    import filecmp
    from typing import Any, Callable, Generator, Iterable, TypeVar

    try:
        # Python 3.8 and 3.9 compatibility
        from typing_extensions import TypeAlias
    except ImportError:
        from typing import TypeAlias  # type: ignore[attr-defined, no-redef]

    from unittest.mock import MagicMock

    from click.testing import Result as ClickInvokeResult
    from git import Commit

    from semantic_release.cli.config import RuntimeContext

    _R = TypeVar("_R")

    GitCommandWrapperType: TypeAlias = Git


def get_func_qual_name(func: Callable) -> str:
    return str.join(".", filter(None, [func.__module__, func.__qualname__]))


def assert_exit_code(
    exit_code: int, result: ClickInvokeResult, cli_cmd: list[str]
) -> bool:
    if result.exit_code == exit_code:
        return True

    raise AssertionError(
        str.join(
            os.linesep,
            [
                f"{result.exit_code} != {exit_code} (actual != expected)",
                "",
                # Explain what command failed
                "Unexpected exit code from command:",
                indent(f"'{str.join(' ', cli_cmd)}'", " " * 2),
            ],
        )
    )


def assert_successful_exit_code(result: ClickInvokeResult, cli_cmd: list[str]) -> bool:
    return assert_exit_code(SUCCESS_EXIT_CODE, result, cli_cmd)


def get_full_qualname(callable_obj: Callable) -> str:
    parts = filter(
        None,
        [
            callable_obj.__module__,
            (
                None
                if callable_obj.__class__.__name__ == "function"
                else callable_obj.__class__.__name__
            ),
            callable_obj.__name__,
        ],
    )
    return str.join(".", parts)


def copy_dir_tree(src_dir: Path | str, dst_dir: Path | str) -> None:
    """Compatibility wrapper for shutil.copytree"""
    # python3.8+
    shutil.copytree(
        src=str(src_dir),
        dst=str(dst_dir),
        dirs_exist_ok=True,
    )


def remove_dir_tree(directory: Path | str = ".", force: bool = False) -> None:
    """
    Compatibility wrapper for shutil.rmtree

    Helpful for deleting directories with .git/* files, which usually have some
    read-only permissions
    """

    def on_read_only_error(_func, path, _exc_info):
        os.chmod(path, stat.S_IWRITE)
        os.unlink(path)

    # Prevent error if already deleted or never existed, that is our desired state
    with suppress(FileNotFoundError):
        shutil.rmtree(str(directory), onerror=on_read_only_error if force else None)


def dynamic_python_import(file_path: Path, module_name: str):
    spec = importlib.util.spec_from_file_location(module_name, str(file_path))
    module = importlib.util.module_from_spec(spec)  # type: ignore[arg-type]
    spec.loader.exec_module(module)  # type: ignore[union-attr]
    return module


@contextmanager
def temporary_working_directory(directory: Path | str) -> Generator[None, None, None]:
    cwd = os.getcwd()
    os.chdir(str(directory))
    try:
        yield
    finally:
        os.chdir(cwd)


def shortuid(length: int = 8) -> str:
    alphabet = string.ascii_lowercase + string.digits

    return "".join(secrets.choice(alphabet) for _ in range(length))


def add_text_to_file(repo: Repo, filename: str, text: str | None = None):
    """Makes a deterministic file change for testing"""
    tgt_file = Path(filename).resolve().absolute()

    # TODO: switch to Path.is_relative_to() when 3.8 support is deprecated
    # if not tgt_file.is_relative_to(Path(repo.working_dir).resolve().absolute()):
    if Path(repo.working_dir).resolve().absolute() not in tgt_file.parents:
        raise ValueError(
            f"File {tgt_file} is not relative to the repository working directory {repo.working_dir}"
        )

    tgt_file.parent.mkdir(parents=True, exist_ok=True)
    file_contents = tgt_file.read_text() if tgt_file.exists() else ""
    line_number = len(file_contents.splitlines())

    file_contents += f"{line_number}  {text or 'default text'}{os.linesep}"
    tgt_file.write_text(file_contents, encoding="utf-8")

    repo.index.add(tgt_file)


def flatten_dircmp(dcmp: filecmp.dircmp) -> list[str]:
    return (
        dcmp.diff_files
        + dcmp.left_only
        + dcmp.right_only
        + [
            os.sep.join((directory, file))
            for directory, cmp in dcmp.subdirs.items()
            for file in flatten_dircmp(cmp)
        ]
    )


def xdist_sort_hack(it: Iterable[_R]) -> Iterable[_R]:
    """
    hack for pytest-xdist

    https://pytest-xdist.readthedocs.io/en/latest/known-limitations.html#workarounds

    taking an iterable of params for a pytest.mark.parametrize decorator, this
    ensures a deterministic sort so that xdist can always work

    Being able to use `pytest -nauto` is a huge speedup on testing
    """
    return dict(enumerate(it)).values()


def actions_output_to_dict(output: str) -> dict[str, str]:
    single_line_var_pattern = regexp(r"^(?P<name>\w+)=(?P<value>.*?)\r?$")
    multiline_var_pattern = regexp(r"^(?P<name>\w+?)<<EOF\r?$")
    multiline_var_pattern_end = regexp(r"^EOF\r?$")

    found_multiline_var = False
    current_var_name = ""
    current_var_value = ""
    result: dict[str, str] = {}
    for line in output.splitlines(keepends=True):
        if found_multiline_var:
            if match := multiline_var_pattern_end.match(line):
                # End of a multiline variable
                found_multiline_var = False
                result[current_var_name] = current_var_value
                continue

            current_var_value += line
            continue

        if match := single_line_var_pattern.match(line):
            # Single line variable
            result[match.group("name")] = match.group("value")
            continue

        if match := multiline_var_pattern.match(line):
            # Start of a multiline variable
            found_multiline_var = True
            current_var_name = match.group("name")
            current_var_value = ""
            continue

    return result


def get_release_history_from_context(runtime_context: RuntimeContext) -> ReleaseHistory:
    with Repo(str(runtime_context.repo_dir)) as git_repo:
        release_history = ReleaseHistory.from_git_history(
            git_repo,
            runtime_context.version_translator,
            runtime_context.commit_parser,
            runtime_context.changelog_excluded_commit_patterns,
        )
    changelog_context = make_changelog_context(
        hvcs_client=runtime_context.hvcs_client,
        release_history=release_history,
        mode=ChangelogMode.INIT,
        prev_changelog_file=Path("CHANGELOG.md"),
        insertion_flag="",
        mask_initial_release=runtime_context.changelog_mask_initial_release,
    )
    changelog_context.bind_to_environment(runtime_context.template_environment)
    return release_history


def prepare_mocked_git_command_wrapper_type(
    **mocked_methods: MagicMock,
) -> type[GitCommandWrapperType]:
    """
    Mock the specified methods of `Repo.GitCommandWrapperType` (`git.Git` by default).

    Initialized `MagicMock` objects are passed as keyword arguments, where the argument
    name is the name of the method to mock.

    For example, the following invocation mocks the `Repo.git.push()` command / method.

    Arrange:
    >>> from unittest.mock import MagicMock
    >>> from git import Repo

    >>> mocked_push = MagicMock()
    >>> cls = prepare_mocked_git_command_wrapper_type(push=mocked_push)
    >>> Repo.GitCommandWrapperType = cls
    >>> repo = Repo(".")

    Act:
    >>> repo.git.push("origin", "master")
    <MagicMock name='mock()' id='...'>

    Assert:
    >>> mocked_push.assert_called_once()
    """

    class MockGitCommandWrapperType(Git):
        def __getattr__(self, name: str) -> Any:
            try:
                return object.__getattribute__(self, f"mocked_{name}")
            except AttributeError:
                return super().__getattr__(name)

    for name, method in mocked_methods.items():
        setattr(MockGitCommandWrapperType, f"mocked_{name}", method)
    return MockGitCommandWrapperType


class CustomParserWithNoOpts(CommitParser[ParseResult, ParserOptions]):
    def parse(self, commit: Commit) -> ParsedCommit | ParseError:
        return ParsedCommit(
            bump=LevelBump.NO_RELEASE,
            type="",
            scope="",
            descriptions=[],
            breaking_descriptions=[],
            commit=commit,
        )


@dataclass
class CustomParserOpts(ParserOptions):
    allowed_tags: Tuple[str, ...] = ("new", "custom")  # noqa: UP006


class CustomParserWithOpts(CommitParser[ParseResult, CustomParserOpts]):
    parser_options = CustomParserOpts

    def parse(self, commit: Commit) -> ParsedCommit | ParseError:
        return ParsedCommit(
            bump=LevelBump.NO_RELEASE,
            type="custom",
            scope="",
            descriptions=[],
            breaking_descriptions=[],
            commit=commit,
        )


class IncompleteCustomParser(CommitParser):
    pass


class CustomConventionalParserWithIgnorePatterns(ConventionalCommitParser):
    def parse(self, commit: Commit) -> ParsedCommit | ParseError:
        if not (parse_msg_result := super().parse_message(str(commit.message))):
            return ParseError(commit, "Unable to parse commit")

        return ParsedCommit.from_parsed_message_result(
            commit,
            ParsedMessageResult(
                **{
                    **parse_msg_result._asdict(),
                    "include_in_changelog": bool(
                        not str(commit.message).startswith("chore")
                    ),
                }
            ),
        )



from __future__ import annotations

import os
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Generator, cast
from unittest import mock

import pytest
import tomlkit

# NOTE: use backport with newer API
from importlib_resources import files

import semantic_release
from semantic_release.cli.config import (
    GlobalCommandLineOptions,
    RawConfig,
    RuntimeContext,
)
from semantic_release.cli.util import load_raw_config_file
from semantic_release.commit_parser import (
    ConventionalCommitParser,
    EmojiCommitParser,
    ScipyCommitParser,
)
from semantic_release.commit_parser.conventional.parser_monorepo import (
    ConventionalCommitMonorepoParser,
)
from semantic_release.hvcs import Bitbucket, Gitea, Github, Gitlab

import tests.conftest
import tests.const
import tests.util
from tests.const import (
    EXAMPLE_CHANGELOG_MD_CONTENT,
    EXAMPLE_CHANGELOG_RST_CONTENT,
    EXAMPLE_PROJECT_NAME,
    EXAMPLE_PROJECT_VERSION,
    EXAMPLE_PYPROJECT_TOML_CONTENT,
    EXAMPLE_RELEASE_NOTES_TEMPLATE,
    EXAMPLE_SETUP_CFG_CONTENT,
    EXAMPLE_SETUP_PY_CONTENT,
)
from tests.util import copy_dir_tree, temporary_working_directory

if TYPE_CHECKING:
    from typing import Any, Protocol, Sequence

    from tomlkit.container import Container as TOMLContainer

    from semantic_release.commit_parser import CommitParser
    from semantic_release.commit_parser._base import ParserOptions
    from semantic_release.commit_parser.token import ParseResult
    from semantic_release.hvcs import HvcsBase
    from semantic_release.version.version import Version

    from tests.conftest import (
        BuildRepoOrCopyCacheFn,
        GetMd5ForSetOfFilesFn,
    )
    from tests.fixtures.git_repo import RepoActions

    ExProjectDir = Path

    class GetWheelFileFn(Protocol):
        def __call__(self, version_str: str, pkg_name: str = ...) -> Path: ...

    class SetFlagFn(Protocol):
        def __call__(self, flag: bool, toml_file: Path | str = ...) -> None: ...

    class UpdatePyprojectTomlFn(Protocol):
        def __call__(
            self, setting: str, value: Any, toml_file: Path | str = ...
        ) -> None: ...

    class UseCustomParserFn(Protocol):
        def __call__(
            self, module_import_str: str, toml_file: Path | str = ...
        ) -> None: ...

    class UseHvcsFn(Protocol):
        def __call__(
            self, domain: str | None = None, toml_file: Path | str = ...
        ) -> type[HvcsBase]: ...

    class UseParserFn(Protocol):
        def __call__(
            self, toml_file: Path | str = ..., monorepo: bool = ...
        ) -> type[CommitParser[ParseResult, ParserOptions]]: ...

    class UseReleaseNotesTemplateFn(Protocol):
        def __call__(self, toml_file: Path | str = ...) -> None: ...

    class UpdateVersionPyFileFn(Protocol):
        def __call__(
            self, version: Version | str, version_file: Path | str = ...
        ) -> None: ...

    class GetHvcsFn(Protocol):
        def __call__(
            self,
            hvcs_client_name: str,
            origin_url: str = ...,
            hvcs_domain: str | None = None,
        ) -> Github | Gitlab | Gitea | Bitbucket: ...

    class ReadConfigFileFn(Protocol):
        """Read the raw config file from `config_path`."""

        def __call__(self, file: Path | str = ...) -> RawConfig: ...

    class LoadRuntimeContextFn(Protocol):
        """Load the runtime context from the config file."""

        def __call__(
            self, cli_opts: GlobalCommandLineOptions | None = None
        ) -> RuntimeContext: ...

    class GetParserFromConfigFileFn(Protocol):
        """Get the commit parser from the config file."""

        def __call__(
            self, file: Path | str = ...
        ) -> CommitParser[ParseResult, ParserOptions]: ...

    class GetExpectedVersionPyFileContentFn(Protocol):
        def __call__(self, version: Version | str) -> str: ...


@pytest.fixture(scope="session")
def deps_files_4_example_project() -> list[Path]:
    return [
        # This file
        Path(__file__).absolute(),
        # because of imports
        Path(tests.const.__file__).absolute(),
        Path(tests.util.__file__).absolute(),
        # because of the fixtures
        Path(tests.conftest.__file__).absolute(),
    ]


@pytest.fixture(scope="session")
def build_spec_hash_4_example_project(
    get_md5_for_set_of_files: GetMd5ForSetOfFilesFn,
    deps_files_4_example_project: list[Path],
) -> str:
    # Generates a hash of the build spec to set when to invalidate the cache
    return get_md5_for_set_of_files(deps_files_4_example_project)


@pytest.fixture(scope="session")
def cached_example_project(
    build_repo_or_copy_cache: BuildRepoOrCopyCacheFn,
    version_py_file: Path,
    pyproject_toml_file: Path,
    setup_cfg_file: Path,
    setup_py_file: Path,
    changelog_md_file: Path,
    changelog_rst_file: Path,
    build_spec_hash_4_example_project: str,
    update_version_py_file: UpdateVersionPyFileFn,
) -> Path:
    """
    Initializes the example project. DO NOT USE DIRECTLY

    Use the `init_example_project` fixture instead.
    """

    def _build_project(cached_project_path: Path) -> Sequence[RepoActions]:
        # purposefully a relative path
        example_dir = version_py_file.parent
        gitignore_contents = dedent(
            f"""
            *.pyc
            /src/**/{version_py_file.name}
            """
        ).lstrip()
        init_py_contents = dedent(
            '''
            """
            An example package with a very informative docstring
            """
            from ._version import __version__


            def hello_world() -> None:
                print("Hello World")
            '''
        ).lstrip()

        with temporary_working_directory(cached_project_path):
            update_version_py_file(EXAMPLE_PROJECT_VERSION)

        file_2_contents: list[tuple[str | Path, str]] = [
            (example_dir / "__init__.py", init_py_contents),
            (".gitignore", gitignore_contents),
            (pyproject_toml_file, EXAMPLE_PYPROJECT_TOML_CONTENT),
            (setup_cfg_file, EXAMPLE_SETUP_CFG_CONTENT),
            (setup_py_file, EXAMPLE_SETUP_PY_CONTENT),
            (changelog_md_file, EXAMPLE_CHANGELOG_MD_CONTENT),
            (changelog_rst_file, EXAMPLE_CHANGELOG_RST_CONTENT),
        ]

        for file, contents in file_2_contents:
            abs_filepath = cached_project_path.joinpath(file).resolve()
            # make sure the parent directory exists
            abs_filepath.parent.mkdir(parents=True, exist_ok=True)
            # write file contents
            abs_filepath.write_text(contents)

        # This is a special build, we don't expose the Repo Actions to the caller
        return []

    # End of _build_project()

    return build_repo_or_copy_cache(
        repo_name=f"project_{EXAMPLE_PROJECT_NAME}",
        build_spec_hash=build_spec_hash_4_example_project,
        build_repo_func=_build_project,
    )


@pytest.fixture
def init_example_project(
    example_project_dir: ExProjectDir,
    cached_example_project: Path,
    change_to_ex_proj_dir: None,
) -> None:
    """This fixture initializes the example project in the current test's project directory."""
    if not cached_example_project.exists():
        raise RuntimeError(
            f"Unable to find cached project files for {EXAMPLE_PROJECT_NAME}"
        )

    # Copy the cached project files into the current test's project directory
    copy_dir_tree(cached_example_project, example_project_dir)


@pytest.fixture
def example_project_with_release_notes_template(
    init_example_project: None,
    use_release_notes_template: UseReleaseNotesTemplateFn,
) -> None:
    use_release_notes_template()


@pytest.fixture(scope="session")
def version_py_file() -> Path:
    return Path("src", EXAMPLE_PROJECT_NAME, "_version.py")


@pytest.fixture(scope="session")
def pyproject_toml_file() -> Path:
    return Path("pyproject.toml")


@pytest.fixture(scope="session")
def setup_cfg_file() -> Path:
    return Path("setup.cfg")


@pytest.fixture(scope="session")
def setup_py_file() -> Path:
    return Path("setup.py")


@pytest.fixture(scope="session")
def dist_dir() -> Path:
    return Path("dist")


@pytest.fixture(scope="session")
def changelog_md_file() -> Path:
    return Path("CHANGELOG.md")


@pytest.fixture(scope="session")
def changelog_rst_file() -> Path:
    return Path("CHANGELOG.rst")


@pytest.fixture(scope="session")
def changelog_template_dir() -> Path:
    return Path("templates")


@pytest.fixture(scope="session")
def default_md_changelog_insertion_flag() -> str:
    return "<!-- version list -->"


@pytest.fixture(scope="session")
def default_rst_changelog_insertion_flag() -> str:
    return f"..{os.linesep}    version list"


@pytest.fixture(scope="session")
def default_changelog_md_template() -> Path:
    """Retrieve the semantic-release default changelog template file"""
    return Path(
        str(
            files(semantic_release.__name__).joinpath(
                Path("data", "templates", "conventional", "md", "CHANGELOG.md.j2")
            )
        )
    ).resolve()


@pytest.fixture(scope="session")
def default_changelog_rst_template() -> Path:
    """Retrieve the semantic-release default changelog template file"""
    return Path(
        str(
            files(semantic_release.__name__).joinpath(
                Path("data", "templates", "conventional", "rst", "CHANGELOG.rst.j2")
            )
        )
    ).resolve()


@pytest.fixture(scope="session")
def get_wheel_file(dist_dir: Path) -> GetWheelFileFn:
    def _get_wheel_file(
        version_str: str,
        pkg_name: str = EXAMPLE_PROJECT_NAME,
    ) -> Path:
        return dist_dir.joinpath(
            f"{pkg_name.replace('-', '_')}-{version_str}-py3-none-any.whl"
        )

    return _get_wheel_file


@pytest.fixture(scope="session")
def read_config_file(pyproject_toml_file: Path) -> ReadConfigFileFn:
    def _read_config_file(file: Path | str = pyproject_toml_file) -> RawConfig:
        config_text = load_raw_config_file(file)
        return RawConfig.model_validate(config_text)

    return _read_config_file


@pytest.fixture(scope="session")
def load_runtime_context(
    read_config_file: ReadConfigFileFn,
    pyproject_toml_file: Path,
) -> LoadRuntimeContextFn:
    def _load_runtime_context(
        cli_opts: GlobalCommandLineOptions | None = None,
    ) -> RuntimeContext:
        opts = cli_opts or GlobalCommandLineOptions(
            config_file=str(pyproject_toml_file),
        )
        raw_config = read_config_file(opts.config_file)
        return RuntimeContext.from_raw_config(raw_config, opts)

    return _load_runtime_context


@pytest.fixture(scope="session")
def get_parser_from_config_file(
    pyproject_toml_file: Path,
    load_runtime_context: LoadRuntimeContextFn,
) -> GetParserFromConfigFileFn:
    def _get_parser_from_config(
        file: Path | str = pyproject_toml_file,
    ) -> CommitParser[ParseResult, ParserOptions]:
        return load_runtime_context(
            cli_opts=GlobalCommandLineOptions(config_file=str(Path(file)))
        ).commit_parser

    return _get_parser_from_config


@pytest.fixture
def example_project_dir(tmp_path: Path) -> ExProjectDir:
    return tmp_path.resolve()


@pytest.fixture
def change_to_ex_proj_dir(
    example_project_dir: ExProjectDir,
) -> Generator[None, None, None]:
    cwd = os.getcwd()
    tgt_dir = str(example_project_dir.resolve())
    if cwd == tgt_dir:
        return

    os.chdir(tgt_dir)
    try:
        yield
    finally:
        os.chdir(cwd)


@pytest.fixture
def use_release_notes_template(
    example_project_template_dir: Path,
    changelog_template_dir: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_file: Path,
) -> UseReleaseNotesTemplateFn:
    def _use_release_notes_template(
        toml_file: Path | str = pyproject_toml_file,
    ) -> None:
        update_pyproject_toml(
            "tool.semantic_release.changelog.template_dir",
            str(changelog_template_dir),
            toml_file=toml_file,
        )
        example_project_template_dir.mkdir(parents=True, exist_ok=True)
        release_notes_j2 = example_project_template_dir / ".release_notes.md.j2"
        release_notes_j2.write_text(EXAMPLE_RELEASE_NOTES_TEMPLATE)

    return _use_release_notes_template


@pytest.fixture
def example_pyproject_toml(
    example_project_dir: ExProjectDir,
    pyproject_toml_file: Path,
) -> Path:
    return example_project_dir / pyproject_toml_file


@pytest.fixture
def example_setup_cfg(
    example_project_dir: ExProjectDir,
    setup_cfg_file: Path,
) -> Path:
    return example_project_dir / setup_cfg_file


@pytest.fixture
def example_setup_py(
    example_project_dir: ExProjectDir,
    setup_py_file: Path,
) -> Path:
    return example_project_dir / setup_py_file


@pytest.fixture
def example_dist_dir(
    example_project_dir: ExProjectDir,
    dist_dir: Path,
) -> Path:
    return example_project_dir / dist_dir


@pytest.fixture
def example_project_wheel_file(
    example_dist_dir: Path,
    get_wheel_file: GetWheelFileFn,
) -> Path:
    return example_dist_dir / get_wheel_file(EXAMPLE_PROJECT_VERSION)


# Note this is just the path and the content may change
@pytest.fixture
def example_changelog_md(
    example_project_dir: ExProjectDir,
    changelog_md_file: Path,
) -> Path:
    return example_project_dir / changelog_md_file


# Note this is just the path and the content may change
@pytest.fixture
def example_changelog_rst(
    example_project_dir: ExProjectDir,
    changelog_rst_file: Path,
) -> Path:
    return example_project_dir / changelog_rst_file


@pytest.fixture
def example_project_template_dir(
    example_project_dir: ExProjectDir,
    changelog_template_dir: Path,
) -> Path:
    return example_project_dir / changelog_template_dir


@pytest.fixture(scope="session")
def get_expected_version_py_file_content() -> GetExpectedVersionPyFileContentFn:
    def _get_expected_version_py_file_content(version: Version | str) -> str:
        return dedent(
            f"""\
            __version__ = "{version}"
            """
        )

    return _get_expected_version_py_file_content


@pytest.fixture(scope="session")
def update_version_py_file(
    version_py_file: Path,
    get_expected_version_py_file_content: GetExpectedVersionPyFileContentFn,
) -> UpdateVersionPyFileFn:
    """
    Updates the specified file with the expected version string content

    :param version: The version to set in the file
    :type version: Version | str

    :param version_file: The file to update
    :type version_file: Path | str
    """

    def _update_version_py_file(
        version: Version | str, version_file: Path | str = version_py_file
    ) -> None:
        cwd_version_py = Path(version_file).resolve()
        cwd_version_py.parent.mkdir(parents=True, exist_ok=True)
        cwd_version_py.write_text(get_expected_version_py_file_content(version))

    return _update_version_py_file


@pytest.fixture(scope="session")
def update_pyproject_toml(pyproject_toml_file: Path) -> UpdatePyprojectTomlFn:
    """Update the pyproject.toml file with the given content."""

    def _update_pyproject_toml(
        setting: str, value: Any, toml_file: Path | str = pyproject_toml_file
    ) -> None:
        cwd_pyproject_toml = Path(toml_file).resolve()
        with open(cwd_pyproject_toml) as rfd:
            pyproject_toml = tomlkit.load(rfd)

        new_setting = {}
        parts = setting.split(".")
        new_setting_key = parts.pop(-1)
        new_setting[new_setting_key] = value

        pointer: TOMLContainer = pyproject_toml
        for part in parts:
            if (next_pointer := pointer.get(part, None)) is None:
                next_pointer = tomlkit.table()
                pointer.add(part, next_pointer)

            pointer = cast("TOMLContainer", next_pointer)

        if value is None:
            pointer.pop(new_setting_key)
        else:
            pointer.update(new_setting)

        with open(cwd_pyproject_toml, "w") as wfd:
            tomlkit.dump(pyproject_toml, wfd)

    return _update_pyproject_toml


@pytest.fixture(scope="session")
def pyproject_toml_config_option_parser() -> str:
    return f"tool.{semantic_release.__name__}.commit_parser"


@pytest.fixture(scope="session")
def pyproject_toml_config_option_remote_type() -> str:
    return f"tool.{semantic_release.__name__}.remote.type"


@pytest.fixture(scope="session")
def pyproject_toml_config_option_remote_domain() -> str:
    return f"tool.{semantic_release.__name__}.remote.domain"


@pytest.fixture(scope="session")
def set_major_on_zero(
    pyproject_toml_file: Path, update_pyproject_toml: UpdatePyprojectTomlFn
) -> SetFlagFn:
    """Turn on/off the major_on_zero setting."""

    def _set_major_on_zero(
        flag: bool, toml_file: Path | str = pyproject_toml_file
    ) -> None:
        update_pyproject_toml("tool.semantic_release.major_on_zero", flag, toml_file)

    return _set_major_on_zero


@pytest.fixture(scope="session")
def set_allow_zero_version(
    pyproject_toml_file: Path, update_pyproject_toml: UpdatePyprojectTomlFn
) -> SetFlagFn:
    """Turn on/off the allow_zero_version setting."""

    def _set_allow_zero_version(
        flag: bool, toml_file: Path | str = pyproject_toml_file
    ) -> None:
        update_pyproject_toml(
            "tool.semantic_release.allow_zero_version", flag, toml_file
        )

    return _set_allow_zero_version


@pytest.fixture(scope="session")
def use_conventional_parser(
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_config_option_parser: str,
) -> UseParserFn:
    """Modify the configuration file to use the Conventional parser."""

    def _use_conventional_parser(
        toml_file: Path | str = pyproject_toml_file, monorepo: bool = False
    ) -> type[CommitParser[ParseResult, ParserOptions]]:
        update_pyproject_toml(
            pyproject_toml_config_option_parser,
            f"conventional{'-monorepo' if monorepo else ''}",
            toml_file=toml_file,
        )
        return cast(
            "type[CommitParser[ParseResult, ParserOptions]]",
            ConventionalCommitMonorepoParser if monorepo else ConventionalCommitParser,
        )

    return _use_conventional_parser


@pytest.fixture(scope="session")
def use_emoji_parser(
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_config_option_parser: str,
) -> UseParserFn:
    """Modify the configuration file to use the Emoji parser."""

    def _use_emoji_parser(
        toml_file: Path | str = pyproject_toml_file, monorepo: bool = False
    ) -> type[CommitParser[ParseResult, ParserOptions]]:
        if monorepo:
            raise ValueError(
                "The Emoji parser does not support monorepo mode. "
                "Use the conventional parser instead."
            )

        update_pyproject_toml(
            pyproject_toml_config_option_parser, "emoji", toml_file=toml_file
        )
        return cast("type[CommitParser[ParseResult, ParserOptions]]", EmojiCommitParser)

    return _use_emoji_parser


@pytest.fixture(scope="session")
def use_scipy_parser(
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_config_option_parser: str,
) -> UseParserFn:
    """Modify the configuration file to use the Scipy parser."""

    def _use_scipy_parser(
        toml_file: Path | str = pyproject_toml_file, monorepo: bool = False
    ) -> type[CommitParser[ParseResult, ParserOptions]]:
        if monorepo:
            raise ValueError(
                "The Scipy parser does not support monorepo mode. "
                "Use the conventional parser instead."
            )

        update_pyproject_toml(
            pyproject_toml_config_option_parser, "scipy", toml_file=toml_file
        )
        return cast("type[CommitParser[ParseResult, ParserOptions]]", ScipyCommitParser)

    return _use_scipy_parser


@pytest.fixture(scope="session")
def use_custom_parser(
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_config_option_parser: str,
) -> UseCustomParserFn:
    """Modify the configuration file to use a user defined string parser."""

    def _use_custom_parser(
        module_import_str: str, toml_file: Path | str = pyproject_toml_file
    ) -> None:
        update_pyproject_toml(
            pyproject_toml_config_option_parser, module_import_str, toml_file=toml_file
        )

    return _use_custom_parser


@pytest.fixture(scope="session")
def get_hvcs(example_git_https_url: str) -> GetHvcsFn:
    hvcs_clients: dict[str, type[HvcsBase]] = {
        "github": Github,
        "gitlab": Gitlab,
        "gitea": Gitea,
        "bitbucket": Bitbucket,
    }

    def _get_hvcs(
        hvcs_client_name: str,
        origin_url: str = example_git_https_url,
        hvcs_domain: str | None = None,
    ) -> Github | Gitlab | Gitea | Bitbucket:
        if (hvcs_class := hvcs_clients.get(hvcs_client_name)) is None:
            raise ValueError(f"Unknown HVCS client name: {hvcs_client_name}")

        # Create HVCS Client instance
        with mock.patch.dict(os.environ, {}, clear=True):
            hvcs = hvcs_class(origin_url, hvcs_domain=hvcs_domain)
            assert hvcs.repo_name  # Force the HVCS client to cache the repo name
            assert hvcs.owner  # Force the HVCS client to cache the owner

        return cast("Github | Gitlab | Gitea | Bitbucket", hvcs)

    return _get_hvcs


@pytest.fixture(scope="session")
def use_github_hvcs(
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_config_option_remote_type: str,
    pyproject_toml_config_option_remote_domain: str,
) -> UseHvcsFn:
    """Modify the configuration file to use GitHub as the HVCS."""

    def _use_github_hvcs(
        domain: str | None = None, toml_file: Path | str = pyproject_toml_file
    ) -> type[HvcsBase]:
        update_pyproject_toml(
            pyproject_toml_config_option_remote_type,
            Github.__name__.lower(),
            toml_file=toml_file,
        )

        if domain is not None:
            update_pyproject_toml(
                pyproject_toml_config_option_remote_domain, domain, toml_file=toml_file
            )

        return Github

    return _use_github_hvcs


@pytest.fixture(scope="session")
def use_gitlab_hvcs(
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_config_option_remote_type: str,
    pyproject_toml_config_option_remote_domain: str,
) -> UseHvcsFn:
    """Modify the configuration file to use GitLab as the HVCS."""

    def _use_gitlab_hvcs(
        domain: str | None = None, toml_file: Path | str = pyproject_toml_file
    ) -> type[HvcsBase]:
        update_pyproject_toml(
            pyproject_toml_config_option_remote_type,
            Gitlab.__name__.lower(),
            toml_file=toml_file,
        )

        if domain is not None:
            update_pyproject_toml(
                pyproject_toml_config_option_remote_domain, domain, toml_file=toml_file
            )

        return Gitlab

    return _use_gitlab_hvcs


@pytest.fixture(scope="session")
def use_gitea_hvcs(
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_config_option_remote_type: str,
    pyproject_toml_config_option_remote_domain: str,
) -> UseHvcsFn:
    """Modify the configuration file to use Gitea as the HVCS."""

    def _use_gitea_hvcs(
        domain: str | None = None, toml_file: Path | str = pyproject_toml_file
    ) -> type[HvcsBase]:
        update_pyproject_toml(
            pyproject_toml_config_option_remote_type,
            Gitea.__name__.lower(),
            toml_file=toml_file,
        )

        if domain is not None:
            update_pyproject_toml(
                pyproject_toml_config_option_remote_domain, domain, toml_file=toml_file
            )

        return Gitea

    return _use_gitea_hvcs


@pytest.fixture(scope="session")
def use_bitbucket_hvcs(
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    pyproject_toml_config_option_remote_type: str,
    pyproject_toml_config_option_remote_domain: str,
) -> UseHvcsFn:
    """Modify the configuration file to use BitBucket as the HVCS."""

    def _use_bitbucket_hvcs(
        domain: str | None = None, toml_file: Path | str = pyproject_toml_file
    ) -> type[HvcsBase]:
        update_pyproject_toml(
            pyproject_toml_config_option_remote_type,
            Bitbucket.__name__.lower(),
            toml_file=toml_file,
        )

        if domain is not None:
            update_pyproject_toml(
                pyproject_toml_config_option_remote_domain, domain, toml_file=toml_file
            )

        return Bitbucket

    return _use_bitbucket_hvcs



from tests.fixtures.repos.trunk_based_dev.repo_w_dual_version_support import *
from tests.fixtures.repos.trunk_based_dev.repo_w_dual_version_support_w_prereleases import *
from tests.fixtures.repos.trunk_based_dev.repo_w_no_tags import *
from tests.fixtures.repos.trunk_based_dev.repo_w_prereleases import *
from tests.fixtures.repos.trunk_based_dev.repo_w_tags import *



# ruff: noqa: T201, allow print statements in non-prod scripts
from __future__ import annotations

from os import getenv
from pathlib import Path
from re import compile as regexp

# Constants
PROJ_DIR = Path(__file__).resolve().parent.parent
DOCS_DIR = PROJ_DIR / "docs"
version_replace_pattern = regexp(r"\$(NEW_VERSION|{NEW_VERSION})")
tag_replace_pattern = regexp(r"\$(NEW_RELEASE_TAG|{NEW_RELEASE_TAG})")


def envsubst(filepath: Path, version: str, release_tag: str) -> None:
    file_content = filepath.read_text()

    found = False
    for pattern, replacement in [
        (version_replace_pattern, version),
        (tag_replace_pattern, release_tag),
    ]:
        if not found and (found := bool(pattern.search(file_content))):
            print(f"Applying envsubst to {filepath}")

        file_content = pattern.sub(replacement, file_content)

    filepath.write_text(file_content)


if __name__ == "__main__":
    new_release_tag = getenv("NEW_RELEASE_TAG")
    new_version = getenv("NEW_VERSION")

    if not new_release_tag:
        print("NEW_RELEASE_TAG environment variable is not set")
        exit(1)

    if not new_version:
        print("NEW_VERSION environment variable is not set")
        exit(1)

    for doc_file in DOCS_DIR.rglob("*.rst"):
        envsubst(filepath=doc_file, version=new_version, release_tag=new_release_tag)



from tests.fixtures.commit_parsers import *
from tests.fixtures.example_project import *
from tests.fixtures.git_repo import *
from tests.fixtures.monorepos import *
from tests.fixtures.repos import *
from tests.fixtures.scipy import *



import pytest

from semantic_release.commit_parser.util import parse_paragraphs


@pytest.mark.parametrize(
    "text, expected",
    [
        ("", []),
        ("\n\n \n\n \n", []),  # Unix (LF) - empty lines
        ("\r\n\r\n \r\n\r\n \n", []),  # Windows (CRLF) - empty lines
        ("\n\nA\n\nB\n", ["A", "B"]),  # Unix (LF)
        ("\r\n\r\nA\r\n\r\nB\n", ["A", "B"]),  # Windows (CRLF)
        (
            "Long\nexplanation\n\nfull of interesting\ndetails",
            ["Long explanation", "full of interesting details"],
        ),
        (
            # Windows uses CRLF
            "Long\r\nexplanation\r\n\r\nfull of interesting\r\ndetails",
            ["Long explanation", "full of interesting details"],
        ),
    ],
)
def test_parse_paragraphs(text, expected):
    assert parse_paragraphs(text) == expected



"""Helper code for interacting with a Gitlab remote VCS"""

from __future__ import annotations

import os
from functools import lru_cache
from pathlib import PurePosixPath
from re import compile as regexp
from typing import TYPE_CHECKING

import gitlab
import gitlab.exceptions
import gitlab.v4
import gitlab.v4.objects
from urllib3.util.url import Url, parse_url

from semantic_release.cli.util import noop_report
from semantic_release.errors import UnexpectedResponse
from semantic_release.globals import logger
from semantic_release.helpers import logged_function
from semantic_release.hvcs.remote_hvcs_base import RemoteHvcsBase
from semantic_release.hvcs.util import suppress_not_found

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any, Callable

    from gitlab.v4.objects import Project as GitLabProject


class Gitlab(RemoteHvcsBase):
    """Gitlab HVCS interface for interacting with Gitlab repositories"""

    DEFAULT_ENV_TOKEN_NAME = "GITLAB_TOKEN"  # noqa: S105
    # purposefully not CI_JOB_TOKEN as it is not a personal access token,
    # It is missing the permission to push to the repository, but has all others (releases, packages, etc.)

    OFFICIAL_NAME = "GitLab"
    DEFAULT_DOMAIN = "gitlab.com"

    def __init__(
        self,
        remote_url: str,
        *,
        hvcs_domain: str | None = None,
        token: str | None = None,
        allow_insecure: bool = False,
        **_kwargs: Any,
    ) -> None:
        super().__init__(remote_url)
        self.token = token
        self.project_namespace = f"{self.owner}/{self.repo_name}"
        self._project: GitLabProject | None = None

        domain_url = self._normalize_url(
            hvcs_domain
            or os.getenv("CI_SERVER_URL", "")
            or f"https://{self.DEFAULT_DOMAIN}",
            allow_insecure=allow_insecure,
        )

        # Strip any auth, query or fragment from the domain
        self._hvcs_domain = parse_url(
            Url(
                scheme=domain_url.scheme,
                host=domain_url.host,
                port=domain_url.port,
                path=str(PurePosixPath(domain_url.path or "/")),
            ).url.rstrip("/")
        )

        self._client = gitlab.Gitlab(self.hvcs_domain.url, private_token=self.token)
        self._api_url = parse_url(self._client.api_url)

    @property
    def project(self) -> GitLabProject:
        if self._project is None:
            self._project = self._client.projects.get(self.project_namespace)
        return self._project

    @lru_cache(maxsize=1)
    def _get_repository_owner_and_name(self) -> tuple[str, str]:
        """
        Get the repository owner and name from GitLab CI environment variables, if
        available, otherwise from parsing the remote url
        """
        if "CI_PROJECT_NAMESPACE" in os.environ and "CI_PROJECT_NAME" in os.environ:
            logger.debug("getting repository owner and name from environment variables")
            return os.environ["CI_PROJECT_NAMESPACE"], os.environ["CI_PROJECT_NAME"]

        return super()._get_repository_owner_and_name()

    @logged_function(logger)
    def create_release(
        self,
        tag: str,
        release_notes: str,
        prerelease: bool = False,  # noqa: ARG002
        assets: list[str] | None = None,  # noqa: ARG002
        noop: bool = False,
    ) -> str:
        """
        Create a release in a remote VCS, adding any release notes and assets to it

        :param tag: The tag to create the release for
        :param release_notes: The changelog description for this version only
        :param prerelease: This parameter has no effect in GitLab
        :param assets: A list of paths to files to upload as assets (TODO: not implemented)
        :param noop: If True, do not perform any actions, only logger intents

        :return: The tag of the release

        :raises: GitlabAuthenticationError: If authentication is not correct
        :raises: GitlabCreateError: If the server cannot perform the request
        """
        if noop:
            noop_report(f"would have created a release for tag {tag}")
            return tag

        logger.info("Creating release for %s", tag)
        # ref: https://docs.gitlab.com/ee/api/releases/index.html#create-a-release
        self.project.releases.create(
            {
                "name": tag,
                "tag_name": tag,
                "tag_message": tag,
                "description": release_notes,
            }
        )
        logger.info("Successfully created release for %s", tag)
        return tag

    @logged_function(logger)
    @suppress_not_found
    def get_release_by_tag(self, tag: str) -> gitlab.v4.objects.ProjectRelease | None:
        """
        Get a release by its tag name.

        :param tag: The tag name to get the release for

        :return: gitlab.v4.objects.ProjectRelease or None if not found

        :raises: gitlab.exceptions.GitlabAuthenticationError: If the user is not authenticated
        """
        try:
            return self.project.releases.get(tag)
        except gitlab.exceptions.GitlabGetError:
            logger.debug("Release %s not found", tag)
            return None
        except KeyError as err:
            raise UnexpectedResponse("JSON response is missing commit.id") from err

    @logged_function(logger)
    def edit_release_notes(  # type: ignore[override]
        self,
        release: gitlab.v4.objects.ProjectRelease,
        release_notes: str,
    ) -> str:
        """
        Update the release notes for a given release.

        :param release: The release object to update
        :param release_notes: The new release notes

        :return: The release id

        :raises: GitlabAuthenticationError: If authentication is not correct
        :raises: GitlabUpdateError: If the server cannot perform the request

        """
        logger.info(
            "Updating release %s [%s]",
            release.name,
            release.attributes.get("commit", {}).get("id"),
        )
        release.description = release_notes
        release.save()
        return str(release.get_id())

    @logged_function(logger)
    def create_or_update_release(
        self, tag: str, release_notes: str, prerelease: bool = False
    ) -> str:
        """
        Create or update a release for the given tag in a remote VCS.

        :param tag: The tag to create or update the release for
        :param release_notes: The changelog description for this version only
        :param prerelease: This parameter has no effect in GitLab

        :return: The release id

        :raises ValueError: If the release could not be created or updated
        :raises gitlab.exceptions.GitlabAuthenticationError: If the user is not authenticated
        :raises GitlabUpdateError: If the server cannot perform the request
        """
        try:
            return self.create_release(
                tag=tag, release_notes=release_notes, prerelease=prerelease
            )
        except gitlab.GitlabCreateError:
            logger.info(
                "New release %s could not be created for project %s",
                tag,
                self.project_namespace,
            )

        if (release_obj := self.get_release_by_tag(tag)) is None:
            raise ValueError(
                f"release for tag {tag} could not be found, and could not be created"
            )

        logger.debug(
            "Found existing release commit %s, updating", release_obj.commit.get("id")
        )
        # If this errors we let it die
        return self.edit_release_notes(
            release=release_obj,
            release_notes=release_notes,
        )

    def remote_url(self, use_token: bool = True) -> str:
        """Get the remote url including the token for authentication if requested"""
        if not (self.token and use_token):
            return self._remote_url

        return self.create_server_url(
            auth=f"gitlab-ci-token:{self.token}",
            path=f"{self.project_namespace}.git",
        )

    def compare_url(self, from_rev: str, to_rev: str) -> str:
        return self.create_repo_url(repo_path=f"/-/compare/{from_rev}...{to_rev}")

    def commit_hash_url(self, commit_hash: str) -> str:
        return self.create_repo_url(repo_path=f"/-/commit/{commit_hash}")

    def issue_url(self, issue_num: str | int) -> str:
        # Strips off any character prefix like '#' that usually exists
        if isinstance(issue_num, str) and (
            match := regexp(r"(\d+)$").search(issue_num)
        ):
            try:
                issue_num = int(match.group(1))
            except ValueError:
                return ""

        if isinstance(issue_num, int):
            return self.create_repo_url(repo_path=f"/-/issues/{issue_num}")

        return ""

    def merge_request_url(self, mr_number: str | int) -> str:
        # Strips off any character prefix like '!' that usually exists
        if isinstance(mr_number, str) and (
            match := regexp(r"(\d+)$").search(mr_number)
        ):
            try:
                mr_number = int(match.group(1))
            except ValueError:
                return ""

        if isinstance(mr_number, int):
            return self.create_repo_url(repo_path=f"/-/merge_requests/{mr_number}")

        return ""

    def pull_request_url(self, pr_number: str | int) -> str:
        return self.merge_request_url(mr_number=pr_number)

    def upload_dists(self, tag: str, dist_glob: str) -> int:
        return super().upload_dists(tag, dist_glob)

    def create_release_url(self, tag: str = "") -> str:
        tag_str = tag.strip()
        return self.create_repo_url(repo_path=f"/-/releases/{tag_str}")

    @staticmethod
    def format_w_official_vcs_name(format_str: str) -> str:
        if "%s" in format_str:
            return format_str % Gitlab.OFFICIAL_NAME

        if "{}" in format_str:
            return format_str.format(Gitlab.OFFICIAL_NAME)

        if "{vcs_name}" in format_str:
            return format_str.format(vcs_name=Gitlab.OFFICIAL_NAME)

        return format_str

    def get_changelog_context_filters(self) -> tuple[Callable[..., Any], ...]:
        return (
            self.create_server_url,
            self.create_repo_url,
            self.commit_hash_url,
            self.compare_url,
            self.issue_url,
            self.merge_request_url,
            self.pull_request_url,
            self.create_release_url,
            self.format_w_official_vcs_name,
        )


RemoteHvcsBase.register(Gitlab)



from __future__ import annotations

import json
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.cli.util import load_raw_config_file, parse_toml
from semantic_release.errors import InvalidConfiguration

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any


@pytest.mark.parametrize(
    "toml_text, expected",
    [
        (
            dedent(
                r"""
                [not_the_right_key]
                foo = "bar"
                """
            ),
            {},
        ),
        (
            dedent(
                r"""
                [semantic_release]
                foo = "bar"
                """
            ),
            {"foo": "bar"},
        ),
        (
            dedent(
                r"""
                [tool.semantic_release]
                abc = 123

                [tool.semantic_release.foo]
                def = 456
                """
            ),
            {"abc": 123, "foo": {"def": 456}},
        ),
    ],
)
def test_parse_toml(toml_text: str, expected: dict[str, Any]):
    assert parse_toml(toml_text) == expected


def test_parse_toml_raises_invalid_configuration_with_invalid_toml():
    invalid_toml = dedent(
        r"""
        [semantic_release]
        foo = bar  # this is not a valid TOML string
        """
    )

    with pytest.raises(InvalidConfiguration):
        parse_toml(invalid_toml)


@pytest.fixture
def raw_toml_config_file(tmp_path: Path) -> Path:
    path = tmp_path / "config.toml"

    path.write_text(
        dedent(
            r"""
            [semantic_release]
            foo = "bar"

            [semantic_release.abc]
            bar = "baz"
            """
        )
    )

    return path


@pytest.fixture
def raw_pyproject_toml_config_file(tmp_path: Path, pyproject_toml_file: Path) -> Path:
    tmp_path.mkdir(exist_ok=True)

    path = tmp_path / pyproject_toml_file

    path.write_text(
        dedent(
            r"""
            [tool.semantic_release]
            foo = "bar"

            [tool.semantic_release.abc]
            bar = "baz"
            """
        )
    )

    return path


@pytest.fixture
def raw_json_config_file(tmp_path: Path) -> Path:
    tmp_path.mkdir(exist_ok=True)

    path = tmp_path / ".releaserc"

    path.write_text(
        json.dumps(
            {"semantic_release": {"foo": "bar", "abc": {"bar": "baz"}}}, indent=4
        )
    )

    return path


@pytest.fixture
def invalid_toml_config_file(tmp_path: Path) -> Path:
    path = tmp_path / "config.toml"

    path.write_text(
        dedent(
            r"""
            [semantic_release]
            foo = bar  # no quotes == invalid

            [semantic_release.abc]
            bar = "baz"
            """
        )
    )

    return path


@pytest.fixture
def invalid_json_config_file(tmp_path: Path) -> Path:
    tmp_path.mkdir(exist_ok=True)

    path = tmp_path / "releaserc.json"

    path.write_text(
        dedent(
            r"""
            {"semantic_release": {foo: "bar", "abc": {bar: "baz"}}}
            """
        )
    )

    return path


@pytest.fixture
def invalid_other_config_file(tmp_path: Path) -> Path:
    # e.g. XML
    path = tmp_path / "config.xml"

    path.write_text(
        dedent(
            r"""
            <semantic_release>
                <foo>bar</foo>
                <abc>
                    <bar>baz</bar>
                </abc>
            </semantic_release>
            """
        )
    )

    return path


@pytest.mark.parametrize(
    "raw_config_file, expected",
    [
        (
            lazy_fixture(raw_toml_config_file.__name__),
            {"foo": "bar", "abc": {"bar": "baz"}},
        ),
        (
            lazy_fixture(raw_pyproject_toml_config_file.__name__),
            {"foo": "bar", "abc": {"bar": "baz"}},
        ),
        (
            lazy_fixture(raw_json_config_file.__name__),
            {"foo": "bar", "abc": {"bar": "baz"}},
        ),
    ],
)
def test_load_raw_config_file_loads_config(
    raw_config_file: Path, expected: dict[str, Any]
):
    assert load_raw_config_file(raw_config_file) == expected


@pytest.mark.parametrize(
    "raw_config_file",
    [
        lazy_fixture(invalid_toml_config_file.__name__),
        lazy_fixture(invalid_json_config_file.__name__),
        lazy_fixture(invalid_other_config_file.__name__),
    ],
)
def test_load_raw_invalid_config_file_raises_error(raw_config_file: Path):
    with pytest.raises(InvalidConfiguration):
        load_raw_config_file(raw_config_file)



from semantic_release.changelog.context import (
    ChangelogContext,
    make_changelog_context,
)
from semantic_release.changelog.release_history import ReleaseHistory
from semantic_release.changelog.template import (
    environment,
    recursive_render,
)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, cast

import pytest
from freezegun import freeze_time

from semantic_release.version.version import Version

from tests.const import RepoActionStep
from tests.fixtures.monorepos.github_flow import (
    monorepo_w_github_flow_w_default_release_channel_conventional_commits,
)
from tests.util import temporary_working_directory

if TYPE_CHECKING:
    from typing import Literal, Sequence
    from unittest.mock import MagicMock

    from requests_mock import Mocker

    from tests.e2e.cmd_version.bump_version.conftest import (
        InitMirrorRepo4RebuildFn,
        RunPSReleaseFn,
    )
    from tests.e2e.conftest import GetSanitizedChangelogContentFn
    from tests.fixtures.example_project import ExProjectDir
    from tests.fixtures.git_repo import (
        BuildRepoFromDefinitionFn,
        BuildSpecificRepoFn,
        CommitConvention,
        GetGitRepo4DirFn,
        RepoActionConfigure,
        RepoActionConfigureMonorepo,
        RepoActionCreateMonorepo,
        RepoActionRelease,
        RepoActions,
        SplitRepoActionsByReleaseTagsFn,
    )


@pytest.mark.parametrize(
    "repo_fixture_name",
    [
        pytest.param(repo_fixture_name, marks=pytest.mark.comprehensive)
        for repo_fixture_name in [
            monorepo_w_github_flow_w_default_release_channel_conventional_commits.__name__,
        ]
    ],
)
def test_githubflow_monorepo_rebuild_1_channel(
    repo_fixture_name: str,
    run_psr_release: RunPSReleaseFn,
    build_monorepo_w_github_flow_w_default_release_channel: BuildSpecificRepoFn,
    split_repo_actions_by_release_tags: SplitRepoActionsByReleaseTagsFn,
    init_mirror_repo_for_rebuild: InitMirrorRepo4RebuildFn,
    example_project_dir: ExProjectDir,
    git_repo_for_directory: GetGitRepo4DirFn,
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    get_sanitized_md_changelog_content: GetSanitizedChangelogContentFn,
    get_sanitized_rst_changelog_content: GetSanitizedChangelogContentFn,
    monorepo_pkg1_pyproject_toml_file: Path,
    monorepo_pkg2_pyproject_toml_file: Path,
    monorepo_pkg1_version_py_file: Path,
    monorepo_pkg2_version_py_file: Path,
    monorepo_pkg1_changelog_md_file: Path,
    monorepo_pkg2_changelog_md_file: Path,
    monorepo_pkg1_changelog_rst_file: Path,
    monorepo_pkg2_changelog_rst_file: Path,
):
    # build target repo into a temporary directory
    target_repo_dir = example_project_dir / repo_fixture_name
    commit_type = cast(
        "CommitConvention", repo_fixture_name.split("commits", 1)[0].split("_")[-2]
    )
    target_repo_definition = build_monorepo_w_github_flow_w_default_release_channel(
        repo_name=repo_fixture_name,
        commit_type=commit_type,
        dest_dir=target_repo_dir,
    )
    target_git_repo = git_repo_for_directory(target_repo_dir)

    # split repo actions by release actions
    release_tags_2_steps = split_repo_actions_by_release_tags(target_repo_definition)

    configuration_steps = cast(
        "Sequence[RepoActionConfigure | RepoActionCreateMonorepo | RepoActionConfigureMonorepo]",
        release_tags_2_steps.pop(None),
    )

    release_versions_2_steps = cast(
        "dict[Version | Literal['Unreleased'], list[RepoActions]]",
        release_tags_2_steps,
    )

    # Create the mirror repo directory
    mirror_repo_dir = init_mirror_repo_for_rebuild(
        mirror_repo_dir=(example_project_dir / "mirror"),
        configuration_steps=configuration_steps,
        files_to_remove=[],
    )

    mirror_git_repo = git_repo_for_directory(mirror_repo_dir)

    # rebuild repo from scratch stopping before each release tag
    for curr_release_key, steps in release_versions_2_steps.items():
        curr_release_str = (
            curr_release_key.as_tag()
            if isinstance(curr_release_key, Version)
            else curr_release_key
        )

        # make sure mocks are clear
        mocked_git_fetch.reset_mock()
        mocked_git_push.reset_mock()
        post_mocker.reset_mock()

        # Extract expected result from target repo
        if curr_release_str != "Unreleased":
            target_git_repo.git.checkout(curr_release_str, detach=True, force=True)

        expected_pkg1_md_changelog_content = get_sanitized_md_changelog_content(
            repo_dir=target_repo_dir, changelog_file=monorepo_pkg1_changelog_md_file
        )
        expected_pkg2_md_changelog_content = get_sanitized_md_changelog_content(
            repo_dir=target_repo_dir, changelog_file=monorepo_pkg2_changelog_md_file
        )
        expected_pkg1_rst_changelog_content = get_sanitized_rst_changelog_content(
            repo_dir=target_repo_dir, changelog_file=monorepo_pkg1_changelog_rst_file
        )
        expected_pkg2_rst_changelog_content = get_sanitized_rst_changelog_content(
            repo_dir=target_repo_dir, changelog_file=monorepo_pkg2_changelog_rst_file
        )
        expected_pkg1_pyproject_toml_content = (
            target_repo_dir / monorepo_pkg1_pyproject_toml_file
        ).read_text()
        expected_pkg2_pyproject_toml_content = (
            target_repo_dir / monorepo_pkg2_pyproject_toml_file
        ).read_text()
        expected_pkg1_version_file_content = (
            target_repo_dir / monorepo_pkg1_version_py_file
        ).read_text()
        expected_pkg2_version_file_content = (
            target_repo_dir / monorepo_pkg2_version_py_file
        ).read_text()
        expected_release_commit_text = target_git_repo.head.commit.message

        # In our repo env, start building the repo from the definition
        build_repo_from_definition(
            dest_dir=mirror_repo_dir,
            # stop before the release step
            repo_construction_steps=steps[
                : -1 if curr_release_str != "Unreleased" else None
            ],
        )

        release_directory = mirror_repo_dir

        for step in steps[::-1]:  # reverse order
            if step["action"] == RepoActionStep.CHANGE_DIRECTORY:
                release_directory = (
                    mirror_repo_dir
                    if str(Path(step["details"]["directory"]))
                    == str(mirror_repo_dir.root)
                    else Path(step["details"]["directory"])
                )

                release_directory = (
                    mirror_repo_dir / release_directory
                    if not release_directory.is_absolute()
                    else release_directory
                )

                if mirror_repo_dir not in release_directory.parents:
                    release_directory = mirror_repo_dir

                break

        # Act: run PSR on the repo instead of the RELEASE step
        if curr_release_str != "Unreleased":
            release_action_step = cast("RepoActionRelease", steps[-1])

            with freeze_time(
                release_action_step["details"]["datetime"]
            ), temporary_working_directory(release_directory):
                run_psr_release(
                    next_version_str=release_action_step["details"]["version"],
                    git_repo=mirror_git_repo,
                    config_toml_path=Path("pyproject.toml"),
                )
        else:
            # run psr changelog command to validate changelog
            pass

        # take measurement after running the version command
        actual_release_commit_text = mirror_git_repo.head.commit.message
        actual_pkg1_pyproject_toml_content = (
            mirror_repo_dir / monorepo_pkg1_pyproject_toml_file
        ).read_text()
        actual_pkg2_pyproject_toml_content = (
            mirror_repo_dir / monorepo_pkg2_pyproject_toml_file
        ).read_text()
        actual_pkg1_version_file_content = (
            mirror_repo_dir / monorepo_pkg1_version_py_file
        ).read_text()
        actual_pkg2_version_file_content = (
            mirror_repo_dir / monorepo_pkg2_version_py_file
        ).read_text()
        actual_pkg1_md_changelog_content = get_sanitized_md_changelog_content(
            repo_dir=mirror_repo_dir, changelog_file=monorepo_pkg1_changelog_md_file
        )
        actual_pkg2_md_changelog_content = get_sanitized_md_changelog_content(
            repo_dir=mirror_repo_dir, changelog_file=monorepo_pkg2_changelog_md_file
        )
        actual_pkg1_rst_changelog_content = get_sanitized_rst_changelog_content(
            repo_dir=mirror_repo_dir, changelog_file=monorepo_pkg1_changelog_rst_file
        )
        actual_pkg2_rst_changelog_content = get_sanitized_rst_changelog_content(
            repo_dir=mirror_repo_dir, changelog_file=monorepo_pkg2_changelog_rst_file
        )

        # Evaluate (normal release actions should have occurred as expected)
        # ------------------------------------------------------------------
        # Make sure version file is updated
        assert (
            expected_pkg1_pyproject_toml_content == actual_pkg1_pyproject_toml_content
        )
        assert (
            expected_pkg2_pyproject_toml_content == actual_pkg2_pyproject_toml_content
        )
        assert expected_pkg1_version_file_content == actual_pkg1_version_file_content
        assert expected_pkg2_version_file_content == actual_pkg2_version_file_content

        # Make sure changelog is updated
        assert expected_pkg1_md_changelog_content == actual_pkg1_md_changelog_content
        assert expected_pkg2_md_changelog_content == actual_pkg2_md_changelog_content
        assert expected_pkg1_rst_changelog_content == actual_pkg1_rst_changelog_content
        assert expected_pkg2_rst_changelog_content == actual_pkg2_rst_changelog_content

        # Make sure commit is created
        assert expected_release_commit_text == actual_release_commit_text

        if curr_release_str != "Unreleased":
            # Make sure tag is created
            assert curr_release_str in [tag.name for tag in mirror_git_repo.tags]

        # Make sure publishing actions occurred
        assert (
            mocked_git_fetch.call_count == 1
        )  # fetch called to check for remote changes
        assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
        assert post_mocker.call_count == 1  # vcs release creation occurred



from typing import Iterable

import pytest

from semantic_release.helpers import ParsedGitUrl, parse_git_url, sort_numerically


@pytest.mark.parametrize(
    "url, expected",
    [
        (
            "http://git.mycompany.com/username/myproject.git",
            ParsedGitUrl("http", "git.mycompany.com", "username", "myproject"),
        ),
        (
            "http://subsubdomain.subdomain.company-net.com/username/myproject.git",
            ParsedGitUrl(
                "http",
                "subsubdomain.subdomain.company-net.com",
                "username",
                "myproject",
            ),
        ),
        (
            "https://github.com/username/myproject.git",
            ParsedGitUrl("https", "github.com", "username", "myproject"),
        ),
        (
            "https://gitlab.com/group/subgroup/myproject.git",
            ParsedGitUrl("https", "gitlab.com", "group/subgroup", "myproject"),
        ),
        (
            "https://git.mycompany.com:4443/username/myproject.git",
            ParsedGitUrl("https", "git.mycompany.com:4443", "username", "myproject"),
        ),
        (
            "https://subsubdomain.subdomain.company-net.com/username/myproject.git",
            ParsedGitUrl(
                "https",
                "subsubdomain.subdomain.company-net.com",
                "username",
                "myproject",
            ),
        ),
        (
            "git://host.xz/path/to/repo.git/",
            ParsedGitUrl("git", "host.xz", "path/to", "repo"),
        ),
        (
            "git://host.xz:9418/path/to/repo.git/",
            ParsedGitUrl("git", "host.xz:9418", "path/to", "repo"),
        ),
        (
            "git@github.com:username/myproject.git",
            ParsedGitUrl("ssh", "git@github.com", "username", "myproject"),
        ),
        (
            "git@subsubdomain.subdomain.company-net.com:username/myproject.git",
            ParsedGitUrl(
                "ssh",
                "git@subsubdomain.subdomain.company-net.com",
                "username",
                "myproject",
            ),
        ),
        (
            "first.last_test-1@subsubdomain.subdomain.company-net.com:username/myproject.git",
            ParsedGitUrl(
                "ssh",
                "first.last_test-1@subsubdomain.subdomain.company-net.com",
                "username",
                "myproject",
            ),
        ),
        (
            "ssh://git@github.com:3759/myproject.git",
            ParsedGitUrl("ssh", "git@github.com", "3759", "myproject"),
        ),
        (
            "ssh://git@github.com:username/myproject.git",
            ParsedGitUrl("ssh", "git@github.com", "username", "myproject"),
        ),
        (
            "ssh://git@bitbucket.org:7999/username/myproject.git",
            ParsedGitUrl("ssh", "git@bitbucket.org:7999", "username", "myproject"),
        ),
        (
            "ssh://git@subsubdomain.subdomain.company-net.com:username/myproject.git",
            ParsedGitUrl(
                "ssh",
                "git@subsubdomain.subdomain.company-net.com",
                "username",
                "myproject",
            ),
        ),
        (
            "git+ssh://git@github.com:username/myproject.git",
            ParsedGitUrl("ssh", "git@github.com", "username", "myproject"),
        ),
        (
            "/Users/username/dev/remote/myproject.git",
            ParsedGitUrl("file", "", "Users/username/dev/remote", "myproject"),
        ),
        (
            "file:///Users/username/dev/remote/myproject.git",
            ParsedGitUrl("file", "", "Users/username/dev/remote", "myproject"),
        ),
        (
            "C:/Users/username/dev/remote/myproject.git",
            ParsedGitUrl("file", "", "C:/Users/username/dev/remote", "myproject"),
        ),
        (
            "file:///C:/Users/username/dev/remote/myproject.git",
            ParsedGitUrl("file", "", "C:/Users/username/dev/remote", "myproject"),
        ),
    ],
)
def test_parse_valid_git_urls(url: str, expected: ParsedGitUrl):
    """Test that a valid given git remote url is parsed correctly."""
    assert expected == parse_git_url(url)


@pytest.mark.parametrize(
    "url",
    [
        "icmp://git",
        "abcdefghijklmnop.git",
        "../relative/path/to/repo.git",
        "http://domain/project.git",
    ],
)
def test_parse_invalid_git_urls(url: str):
    """Test that an invalid git remote url throws a ValueError."""
    with pytest.raises(ValueError):
        parse_git_url(url)


@pytest.mark.parametrize(
    "unsorted_list, sorted_list, reverse, allow_hex",
    [
        pytest.param(
            unsorted_list,
            sorted_list,
            reverse,
            allow_hex,
            id=f"({i}) {test_id}",
        )
        for i, (test_id, unsorted_list, sorted_list, reverse, allow_hex) in enumerate(
            [
                (
                    "Only numbers (with mixed digits, ASC)",
                    ["5", "3", "10"],
                    ["3", "5", "10"],
                    False,
                    False,
                ),
                (
                    "Only numbers (with mixed digits, DESC)",
                    ["5", "3", "10"],
                    ["10", "5", "3"],
                    True,
                    False,
                ),
                (
                    "Only PR numbers (ASC)",
                    ["#5", "#3", "#10"],
                    ["#3", "#5", "#10"],
                    False,
                    False,
                ),
                (
                    "Only PR numbers (DESC)",
                    ["#5", "#3", "#10"],
                    ["#10", "#5", "#3"],
                    True,
                    False,
                ),
                (
                    "Multiple prefixes (ASC)",
                    ["#5", "PR#3", "PR#10", "#100"],
                    ["#5", "#100", "PR#3", "PR#10"],
                    False,
                    False,
                ),
                (
                    "Multiple prefixes (DESC)",
                    ["#5", "PR#3", "PR#10", "#100"],
                    ["#100", "#5", "PR#10", "PR#3"],
                    True,
                    False,
                ),
                (
                    "No numbers mixed with mulitple prefixes (ASC)",
                    ["word", "#100", "#1000", "PR#45"],
                    ["#100", "#1000", "PR#45", "word"],
                    False,
                    False,
                ),
                (
                    "No numbers mixed with mulitple prefixes (DESC)",
                    ["word", "#100", "#1000", "PR#45"],
                    ["#1000", "#100", "PR#45", "word"],
                    True,
                    False,
                ),
                (
                    "Commit hash links in RST link format (ASC)",
                    [".. _8ab43ed:", ".. _7ffed34:", ".. _a3b4c54:"],
                    [".. _7ffed34:", ".. _8ab43ed:", ".. _a3b4c54:"],
                    False,
                    True,
                ),
                (
                    "Commit hash links in RST link format (DESC)",
                    [".. _8ab43ed:", ".. _7ffed34:", ".. _a3b4c54:"],
                    [".. _a3b4c54:", ".. _8ab43ed:", ".. _7ffed34:"],
                    True,
                    True,
                ),
                (
                    "Mixed numbers, PR numbers, and commit hash links in RST link format (ASC)",
                    [
                        ".. _#5:",
                        ".. _8ab43ed:",
                        ".. _PR#3:",
                        ".. _#20:",
                        ".. _7ffed34:",
                        ".. _#100:",
                        ".. _a3b4c54:",
                    ],
                    [
                        ".. _7ffed34:",
                        ".. _8ab43ed:",
                        ".. _a3b4c54:",
                        ".. _#5:",
                        ".. _#20:",
                        ".. _#100:",
                        ".. _PR#3:",
                    ],
                    False,
                    True,
                ),
                (
                    "Mixed numbers, PR numbers, and commit hash links in RST link format (DESC)",
                    [
                        ".. _#5:",
                        ".. _8ab43ed:",
                        ".. _PR#3:",
                        ".. _#20:",
                        ".. _7ffed34:",
                        ".. _#100:",
                        ".. _a3b4c54:",
                    ],
                    [
                        ".. _a3b4c54:",
                        ".. _8ab43ed:",
                        ".. _7ffed34:",
                        ".. _#100:",
                        ".. _#20:",
                        ".. _#5:",
                        ".. _PR#3:",
                    ],
                    True,
                    True,
                ),
                (
                    # No change since the prefixes are always alphabetical, asc/desc only is b/w numbers
                    "Same numbers with different prefixes (ASC)",
                    ["PR#5", "#5"],
                    ["#5", "PR#5"],
                    False,
                    False,
                ),
                (
                    "Same numbers with different prefixes (DESC)",
                    ["#5", "PR#5"],
                    ["#5", "PR#5"],
                    True,
                    False,
                ),
            ],
            start=1,
        )
    ],
)
def test_sort_numerically(
    unsorted_list: Iterable[str],
    sorted_list: Iterable[str],
    reverse: bool,
    allow_hex: bool,
):
    actual_list = sort_numerically(
        iterable=unsorted_list,
        reverse=reverse,
        allow_hex=allow_hex,
    )
    assert sorted_list == actual_list



from __future__ import annotations

from itertools import zip_longest
from typing import Tuple

from pydantic.dataclasses import dataclass

from semantic_release.commit_parser._base import ParserOptions
from semantic_release.enums import LevelBump


@dataclass
class ConventionalCommitParserOptions(ParserOptions):
    """Options dataclass for the ConventionalCommitParser."""

    minor_tags: Tuple[str, ...] = ("feat",)
    """Commit-type prefixes that should result in a minor release bump."""

    patch_tags: Tuple[str, ...] = ("fix", "perf")
    """Commit-type prefixes that should result in a patch release bump."""

    other_allowed_tags: Tuple[str, ...] = (
        "build",
        "chore",
        "ci",
        "docs",
        "style",
        "refactor",
        "test",
    )
    """Commit-type prefixes that are allowed but do not result in a version bump."""

    allowed_tags: Tuple[str, ...] = (
        *minor_tags,
        *patch_tags,
        *other_allowed_tags,
    )
    """
    All commit-type prefixes that are allowed.

    These are used to identify a valid commit message. If a commit message does not start with
    one of these prefixes, it will not be considered a valid commit message.
    """

    default_bump_level: LevelBump = LevelBump.NO_RELEASE
    """The minimum bump level to apply to valid commit message."""

    parse_squash_commits: bool = True
    """Toggle flag for whether or not to parse squash commits"""

    ignore_merge_commits: bool = True
    """Toggle flag for whether or not to ignore merge commits"""

    @property
    def tag_to_level(self) -> dict[str, LevelBump]:
        """A mapping of commit tags to the level bump they should result in."""
        return self._tag_to_level

    def __post_init__(self) -> None:
        self._tag_to_level: dict[str, LevelBump] = {
            str(tag): level
            for tag, level in [
                # we have to do a type ignore as zip_longest provides a type that is not specific enough
                # for our expected output. Due to the empty second array, we know the first is always longest
                # and that means no values in the first entry of the tuples will ever be a LevelBump. We
                # apply a str() to make mypy happy although it will never happen.
                *zip_longest(self.allowed_tags, (), fillvalue=self.default_bump_level),
                *zip_longest(self.patch_tags, (), fillvalue=LevelBump.PATCH),
                *zip_longest(self.minor_tags, (), fillvalue=LevelBump.MINOR),
            ]
            if "|" not in str(tag)
        }



from tests.fixtures.monorepos.example_monorepo import *
from tests.fixtures.monorepos.git_monorepo import *
from tests.fixtures.monorepos.github_flow import *
from tests.fixtures.monorepos.trunk_based_dev import *



from __future__ import annotations

# TODO: Remove v11
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

from deprecated.sphinx import deprecated

from semantic_release.globals import logger
from semantic_release.version.declarations.enum import VersionStampType
from semantic_release.version.declarations.i_version_replacer import IVersionReplacer
from semantic_release.version.declarations.pattern import PatternVersionDeclaration
from semantic_release.version.declarations.toml import TomlVersionDeclaration

if TYPE_CHECKING:  # pragma: no cover
    from semantic_release.version.version import Version


# Globals
__all__ = [
    "IVersionReplacer",
    "VersionStampType",
    "PatternVersionDeclaration",
    "TomlVersionDeclaration",
    "VersionDeclarationABC",
]


@deprecated(
    version="9.20.0",
    reason=str.join(
        " ",
        [
            "Refactored to composition paradigm using the new IVersionReplacer interface.",
            "This class will be removed in a future release",
        ],
    ),
)
class VersionDeclarationABC(ABC):
    """
    ABC for classes representing a location in which a version is declared somewhere
    within the source tree of the repository
    """

    def __init__(self, path: Path | str, search_text: str) -> None:
        self.path = Path(path)
        if not self.path.exists():
            raise FileNotFoundError(f"path {self.path.resolve()!r} does not exist")
        self.search_text = search_text
        self._content: str | None = None

    @property
    def content(self) -> str:
        """
        The content of the source file in which the version is stored. This property
        is cached in the instance variable _content
        """
        if self._content is None:
            logger.debug(
                "No content stored, reading from source file %s", self.path.resolve()
            )
            self._content = self.path.read_text()
        return self._content

    @content.deleter
    def content(self) -> None:
        logger.debug("resetting instance-stored source file contents")
        self._content = None

    @abstractmethod
    def parse(self) -> set[Version]:
        """
        Return a set of the versions which can be parsed from the file.
        Because a source can match in multiple places, this method returns a
        set of matches. Generally, there should only be one element in this
        set (i.e. even if the version is specified in multiple places, it
        should be the same version in each place), but enforcing that condition
        is not mandatory or expected.
        """

    @abstractmethod
    def replace(self, new_version: Version) -> str:
        """
        Update the versions.
        This method reads the underlying file, replaces each occurrence of the
        matched pattern, then writes the updated file.
        :param new_version: The new version number as a `Version` instance
        """

    def write(self, content: str) -> None:
        r"""
        Write new content back to the source path.
        Use alongside .replace():
        >>> class MyVD(VersionDeclarationABC):
        ...     def parse(self): ...
        ...     def replace(self, new_version: Version): ...
        ...     def write(self, content: str): ...

        >>> new_version = Version.parse("1.2.3")
        >>> vd = MyVD("path", r"__version__ = (?P<version>\d+\d+\d+)")
        >>> vd.write(vd.replace(new_version))
        """
        logger.debug("writing content to %r", self.path.resolve())
        self.path.write_text(content)
        self._content = None



from __future__ import annotations

from pathlib import Path
from re import compile as regexp, error as RegExpError  # noqa: N812
from typing import TYPE_CHECKING, Any, Iterable, Tuple

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass

# typing_extensions is for Python 3.8, 3.9, 3.10 compatibility
from typing_extensions import Annotated

from semantic_release.commit_parser.conventional.options import (
    ConventionalCommitParserOptions,
)

if TYPE_CHECKING:  # pragma: no cover
    pass


@dataclass
class ConventionalCommitMonorepoParserOptions(ConventionalCommitParserOptions):
    # TODO: add example into the docstring
    """Options dataclass for ConventionalCommitMonorepoParser."""

    path_filters: Annotated[Tuple[str, ...], Field(validate_default=True)] = (".",)
    """
    A set of relative paths to filter commits by. Only commits with file changes that
    match these file paths or its subdirectories will be considered valid commits.

    Syntax is similar to .gitignore with file path globs and inverse file match globs
    via `!` prefix. Paths should be relative to the current working directory.
    """

    scope_prefix: str = ""
    """
    A prefix that will be striped from the scope when parsing commit messages.

    If set, it will cause unscoped commits to be ignored. Use this in tandem with
    the `path_filters` option to filter commits by directory and scope. This will
    be fed into a regular expression so you must escape any special characters that
    are meaningful in regular expressions (e.g. `.`, `*`, `?`, `+`, etc.) if you want
    to match them literally.
    """

    @field_validator("path_filters", mode="before")
    @classmethod
    def convert_strs_to_paths(cls, value: Any) -> tuple[str, ...]:
        if isinstance(value, str):
            return (value,)

        if isinstance(value, Path):
            return (str(value),)

        if isinstance(value, Iterable):
            results: list[str] = []
            for val in value:
                if isinstance(val, (str, Path)):
                    results.append(str(Path(val)))
                    continue

                msg = f"Invalid type: {type(val)}, expected str or Path."
                raise TypeError(msg)

            return tuple(results)

        msg = f"Invalid type: {type(value)}, expected str, Path, or Iterable."
        raise TypeError(msg)

    @field_validator("path_filters", mode="after")
    @classmethod
    def resolve_path(cls, dir_path_strs: tuple[str, ...]) -> tuple[str, ...]:
        return tuple(
            (
                f"!{Path(str_path[1:]).expanduser().absolute().resolve()}"
                # maintains the negation prefix if it exists
                if str_path.startswith("!")
                # otherwise, resolve the path normally
                else str(Path(str_path).expanduser().absolute().resolve())
            )
            for str_path in dir_path_strs
        )

    @field_validator("scope_prefix", mode="after")
    @classmethod
    def validate_scope_prefix(cls, scope_prefix: str) -> str:
        if not scope_prefix:
            return ""

        # Allow the special case of a plain wildcard although it's not a valid regex
        if scope_prefix == "*":
            return ".*"

        try:
            regexp(scope_prefix)
        except RegExpError as err:
            raise ValueError(f"Invalid regex {scope_prefix!r}") from err

        return scope_prefix



from semantic_release.commit_parser._base import (
    CommitParser,
    ParserOptions,
)
from semantic_release.commit_parser.angular import (
    AngularCommitParser,
    AngularParserOptions,
)
from semantic_release.commit_parser.conventional import (
    ConventionalCommitMonorepoParser,
    ConventionalCommitMonorepoParserOptions,
    ConventionalCommitParser,
    ConventionalCommitParserOptions,
)
from semantic_release.commit_parser.emoji import (
    EmojiCommitParser,
    EmojiParserOptions,
)
from semantic_release.commit_parser.scipy import (
    ScipyCommitParser,
    ScipyParserOptions,
)
from semantic_release.commit_parser.tag import (
    TagCommitParser,
    TagParserOptions,
)
from semantic_release.commit_parser.token import (
    ParsedCommit,
    ParseError,
    ParseResult,
    ParseResultType,
)

__all__ = [
    "CommitParser",
    "ParserOptions",
    "AngularCommitParser",
    "AngularParserOptions",
    "ConventionalCommitParser",
    "ConventionalCommitParserOptions",
    "ConventionalCommitMonorepoParser",
    "ConventionalCommitMonorepoParserOptions",
    "EmojiCommitParser",
    "EmojiParserOptions",
    "ScipyCommitParser",
    "ScipyParserOptions",
    "TagCommitParser",
    "TagParserOptions",
    "ParsedCommit",
    "ParseError",
    "ParseResult",
    "ParseResultType",
]



"""Common functionality and interface for interacting with Git remote VCS"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from pathlib import PurePosixPath
from typing import TYPE_CHECKING

from urllib3.util.url import Url, parse_url

from semantic_release.hvcs import HvcsBase

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any


class RemoteHvcsBase(HvcsBase, metaclass=ABCMeta):
    """
    Interface for subclasses interacting with a remote VCS

    This abstract class is defined to provide common helper functions and
    a set of basic methods that all remote VCS environments usually support.

    If the remote vcs implementation (via subclass) does not support a functionality
    then it can just call super()'s method which defaults as a non-supported log
    message and empty results.  This is more straightforward than checking for
    NotImplemented around every function call in the core library code.
    """

    DEFAULT_ENV_TOKEN_NAME = "HVCS_TOKEN"  # noqa: S105

    def __init__(self, remote_url: str, *_args: Any, **_kwargs: Any) -> None:
        super().__init__(remote_url)
        self._hvcs_domain: Url | None = None
        self._api_url: Url | None = None

    @property
    def hvcs_domain(self) -> Url:
        if self._hvcs_domain is None:
            raise RuntimeError("Property 'hvcs_domain' was used before it was set!")
        return self._hvcs_domain

    @property
    def api_url(self) -> Url:
        if self._api_url is None:
            raise RuntimeError("Property 'api_url' was used before it was set!")
        return self._api_url

    @abstractmethod
    def upload_dists(self, tag: str, dist_glob: str) -> int:
        """
        Upload built distributions to a release on a remote VCS that
        supports such uploads
        """
        self._not_supported(self.upload_dists.__name__)
        return 0

    @abstractmethod
    def create_release(
        self,
        tag: str,
        release_notes: str,
        prerelease: bool = False,
        assets: list[str] | None = None,
        noop: bool = False,
    ) -> int | str:
        """
        Create a release in a remote VCS, if supported

        Which includes uploading any assets as part of the release
        """
        self._not_supported(self.create_release.__name__)
        return -1

    @abstractmethod
    def create_or_update_release(
        self, tag: str, release_notes: str, prerelease: bool = False
    ) -> int | str:
        """
        Create or update a release for the given tag in a remote VCS, attaching the
        given changelog, if supported
        """
        self._not_supported(self.create_or_update_release.__name__)
        return -1

    def create_server_url(
        self,
        path: str,
        auth: str | None = None,
        query: str | None = None,
        fragment: str | None = None,
    ) -> str:
        # Ensure any path prefix is transferred but not doubled up on the derived url
        normalized_path = (
            f"{self.hvcs_domain.path}/{path}"
            if self.hvcs_domain.path and not path.startswith(self.hvcs_domain.path)
            else path
        )
        return self._derive_url(
            self.hvcs_domain,
            path=normalized_path,
            auth=auth,
            query=query,
            fragment=fragment,
        )

    def create_repo_url(
        self,
        repo_path: str,
        query: str | None = None,
        fragment: str | None = None,
    ) -> str:
        return self.create_server_url(
            path=f"/{self.owner}/{self.repo_name}/{repo_path}",
            query=query,
            fragment=fragment,
        )

    def create_api_url(
        self,
        endpoint: str,
        auth: str | None = None,
        query: str | None = None,
        fragment: str | None = None,
    ) -> str:
        # Ensure any api path prefix is transferred but not doubled up on the derived api url
        normalized_endpoint = (
            f"{self.api_url.path}/{endpoint}"
            if self.api_url.path and not endpoint.startswith(self.api_url.path)
            else endpoint
        )
        return self._derive_url(
            self.api_url,
            path=normalized_endpoint,
            auth=auth,
            query=query,
            fragment=fragment,
        )

    @staticmethod
    def _derive_url(
        base_url: Url,
        path: str,
        auth: str | None = None,
        query: str | None = None,
        fragment: str | None = None,
    ) -> str:
        overrides = dict(
            filter(
                lambda x: x[1] is not None,
                {
                    "auth": auth,
                    "path": str(PurePosixPath("/", path.lstrip("/"))),
                    "query": query,
                    "fragment": fragment,
                }.items(),
            )
        )
        return Url(
            **{
                **base_url._asdict(),
                **overrides,
            }
        ).url.rstrip("/")

    @staticmethod
    def _validate_url_scheme(url: Url, allow_insecure: bool = False) -> None:
        if url.scheme == "http" and not allow_insecure:
            raise ValueError("Insecure connections are currently disabled.")

        if url.scheme not in ["http", "https"]:
            raise ValueError(
                f"Invalid scheme {url.scheme} for {url.host}. "
                "Only http and https are supported."
            )

    @staticmethod
    def _normalize_url(url: Url | str, allow_insecure: bool = False) -> Url:
        """
        Function to ensure url scheme is populated & allowed

        Raises
        ------
        TypeError: when url parameter is not a string or parsable url
        ValueError: when the url scheme is not http or https

        """
        tgt_url = parse_url(url) if isinstance(url, str) else url
        if not isinstance(tgt_url, Url):
            raise TypeError(
                f"Invalid url type ({type(tgt_url)}) received, expected Url or string"
            )

        if not tgt_url.scheme:
            new_scheme = "http" if allow_insecure else "https"
            tgt_url = Url(**{**tgt_url._asdict(), "scheme": new_scheme})

        RemoteHvcsBase._validate_url_scheme(tgt_url, allow_insecure=allow_insecure)
        return tgt_url



import operator
import random

import pytest

from semantic_release.enums import LevelBump
from semantic_release.errors import InvalidVersion
from semantic_release.version.version import Version

random.seed(0)

EXAMPLE_VERSION_STRINGS = [
    "1.0.0",
    "0.1.0",
    "0.0.1",
    "1.2.3",
    "0.2.4",
    "2.6.15",
    "13.0.0",
    "4.26.3",
    "1.0.0-rc.1",
    "4.26.0-beta.3",
    "5.3.1+local.123456",
    "9.22.0-alpha.4+build.9999",
]


@pytest.mark.parametrize(
    "version_parts",
    # Major, minor, patch, prerelease_token, prerelease_revision, build_metadata
    [
        (1, 0, 0, "rc", None, ""),
        (0, 1, 0, "rc", None, ""),
        (0, 0, 1, "rc", None, ""),
        (1, 2, 3, "rc", None, ""),
        (0, 2, 4, "rc", None, ""),
        (2, 6, 15, "rc", None, ""),
        (13, 0, 0, "rc", None, ""),
        (4, 26, 3, "rc", None, ""),
        (1, 0, 0, "rc", 1, ""),
        (4, 26, 3, "beta", 3, ""),
        (5, 3, 1, "rc", None, "local.123456"),
        (9, 22, 0, "alpha", 4, "build.9999"),
        (17, 0, 3, "custom-token", 12, ""),
        (17, 0, 3, "custom-token-3-6-9", 12, ""),
        (17, 0, 3, "custom-token", 12, "build.9999"),
    ],
)
def test_version_parse_succeeds(version_parts):
    full = f"{version_parts[0]}.{version_parts[1]}.{version_parts[2]}"
    prerelease = f"-{version_parts[3]}.{version_parts[4]}" if version_parts[4] else ""
    build_metadata = f"+{version_parts[5]}" if version_parts[5] else ""
    version_str = f"{full}{prerelease}{build_metadata}"

    version = Version.parse(version_str)
    assert version.major == version_parts[0]
    assert version.minor == version_parts[1]
    assert version.patch == version_parts[2]
    assert version.prerelease_token == version_parts[3]
    assert version.prerelease_revision == version_parts[4]
    assert version.build_metadata == version_parts[5]
    assert str(version) == version_str


@pytest.mark.parametrize(
    "bad_version",
    [
        "v1.2.3",
        "2.3",
        "2.1.dev0",
        "2.1.4.post5",
        "alpha-1.2.3",
        "17.0.3-custom_token.12",
        "9",
        "4.1.2!-major",
        "%.*.?",
        "M2.m3.p1",
    ],
)
def test_version_parse_fails(bad_version):
    with pytest.raises(InvalidVersion, match=f"{bad_version!r}"):
        Version.parse(bad_version)


@pytest.fixture(params=EXAMPLE_VERSION_STRINGS)
def a_version(request):
    return Version.parse(request.param)


@pytest.mark.parametrize(
    "bad_format", ["non_unique_format", "case_sensitive_{Version}", "typo_{versione}"]
)
def test_tag_format_must_contain_version_field(a_version, bad_format):
    with pytest.raises(ValueError, match=f"Invalid tag_format {bad_format!r}"):
        a_version.tag_format = bad_format


@pytest.mark.parametrize(
    "tag_format",
    [
        "v{version}",
        "dev-{version}",
        "release-_-{version}",
        "{version}-final",
        "{version}-demo-{version}",
    ],
)
def test_change_tag_format_updates_as_tag_method(a_version, tag_format):
    a_version.tag_format = tag_format
    assert a_version.as_tag() == tag_format.format(version=str(a_version))


@pytest.mark.parametrize(
    "version_str, is_prerelease",
    [
        ("1.0.0", False),
        ("14.33.10", False),
        ("2.1.1-rc.1", True),
        ("65.1.2-alpha.4", True),
        ("17.0.3-custom-token.12", True),
        ("17.0.3-custom-token.12+20220101000000", True),
        ("4.2.4+zzzz9000", False),
    ],
)
def test_version_prerelease(version_str, is_prerelease):
    assert Version.parse(version_str).is_prerelease == is_prerelease


def test_version_eq_succeeds(a_version):
    assert a_version == a_version
    assert a_version == str(a_version)


@pytest.mark.parametrize(
    "lower_version, upper_version",
    [
        ("1.0.0", "1.0.1"),
        ("1.0.0", "1.1.0"),
        ("1.0.0", "1.1.1"),
        ("1.0.0", "2.0.0"),
        ("1.0.0-rc.1", "1.0.0"),
        ("1.0.0-rc.1", "1.0.0-rc.2"),
        ("1.0.0-alpha.1", "1.0.1-beta.1"),
        ("1.0.1", "2.0.0-rc.1"),
    ],
)
@pytest.mark.parametrize(
    "op",
    [
        operator.lt,
        operator.le,
        operator.ne,
        lambda left, right: left < right,
        lambda left, right: left <= right,
        lambda left, right: left != right,
    ],
)
def test_version_comparator_succeeds(lower_version, upper_version, op):
    left = Version.parse(lower_version)
    right = Version.parse(upper_version)
    # Test both on Version $op string and on Version $op Version
    assert op(left, right)
    assert op(left, str(right))


@pytest.mark.parametrize(
    "bad_input",
    [
        5,
        "foo-4.22",
        ["a", list, "of", 5, ("things",)],
        (1, 2, 3),
        {"foo": 12},
        "v2.3.4",
    ],
)
@pytest.mark.parametrize(
    "op",
    [
        operator.lt,
        operator.le,
        operator.gt,
        operator.ge,
    ],
)
def test_version_comparator_typeerror(bad_input, op):
    with pytest.raises(TypeError):
        op(Version.parse("1.4.5"), bad_input)


def test_version_equality(a_version):
    assert a_version == Version.parse(str(a_version))


@pytest.mark.parametrize(
    "left, right", [("1.2.3+local.3", "1.2.3"), ("2.1.1-rc.1+build.7777", "2.1.1-rc.1")]
)
def test_version_equality_when_build_metadata_lost(left, right):
    assert Version.parse(left) == Version.parse(right)


@pytest.mark.parametrize(
    "lower_version, upper_version, level",
    [
        ("1.0.0", "1.0.1", LevelBump.PATCH),
        ("1.0.0", "1.1.0", LevelBump.MINOR),
        ("1.0.0", "1.1.1", LevelBump.MINOR),
        ("1.0.0", "2.0.0", LevelBump.MAJOR),
        ("1.0.0-rc.1", "1.0.0", LevelBump.PRERELEASE_REVISION),
        ("1.0.1", "1.1.0-rc.1", LevelBump.MINOR),
        ("1.0.0-rc.1", "1.0.0-rc.2", LevelBump.PRERELEASE_REVISION),
        ("1.0.0-alpha.1", "1.0.1-beta.1", LevelBump.PATCH),
        ("1.0.1", "2.0.0-rc.1", LevelBump.MAJOR),
    ],
)
def test_version_difference(lower_version, upper_version, level):
    left = Version.parse(lower_version)
    right = Version.parse(upper_version)
    assert (left - right) is level
    assert (right - left) is level


@pytest.mark.parametrize(
    "bad_input",
    [
        5,
        "foo-4.22",
        ["a", list, "of", 5, ("things",)],
        (1, 2, 3),
        {"foo": 12},
        "v2.3.4",
    ],
)
def test_unimplemented_version_diff(bad_input):
    with pytest.raises(TypeError, match=r"unsupported operand type"):
        Version.parse("1.2.3") - bad_input


@pytest.mark.parametrize(
    "current_version, prerelease_token, expected_prerelease_version",
    [
        ("1.2.3", "rc", "1.2.3-rc.1"),
        ("1.1.1-rc.2", "rc", "1.1.1-rc.2"),
        ("2.0.0", "beta", "2.0.0-beta.1"),
        ("2.0.0-beta.1", "beta", "2.0.0-beta.1"),
    ],
)
def test_version_to_prerelease_defaults(
    current_version, prerelease_token, expected_prerelease_version
):
    assert Version.parse(current_version).to_prerelease(
        token=prerelease_token
    ) == Version.parse(expected_prerelease_version)


@pytest.mark.parametrize(
    "current_version, prerelease_token, revision, expected_prerelease_version",
    [
        ("1.2.3", "rc", 3, "1.2.3-rc.3"),
        ("1.1.1-rc.1", "rc", 3, "1.1.1-rc.3"),
        ("2.0.0", "beta", None, "2.0.0-beta.1"),
        ("2.0.0-beta.1", "beta", 4, "2.0.0-beta.4"),
    ],
)
def test_version_to_prerelease_with_params(
    current_version, prerelease_token, revision, expected_prerelease_version
):
    assert Version.parse(current_version).to_prerelease(
        token=prerelease_token, revision=revision
    ) == Version.parse(expected_prerelease_version)


@pytest.mark.parametrize(
    "current_version, expected_final_version",
    [
        ("1.2.3-rc.1", "1.2.3"),
        ("1.2.3", "1.2.3"),
        ("1.1.1-rc.2", "1.1.1"),
        ("2.0.0-beta.1", "2.0.0"),
        ("2.27.0", "2.27.0"),
    ],
)
def test_version_finalize_version(current_version, expected_final_version):
    v1 = Version.parse(current_version)
    assert v1.finalize_version() == Version.parse(
        expected_final_version, prerelease_token=v1.prerelease_token
    )


@pytest.mark.parametrize(
    "current_version, level, new_version",
    [
        ("1.2.3", LevelBump.NO_RELEASE, "1.2.3"),
        ("1.2.3", LevelBump.PRERELEASE_REVISION, "1.2.3-rc.1"),
        ("1.2.3", LevelBump.PATCH, "1.2.4"),
        ("1.2.3", LevelBump.MINOR, "1.3.0"),
        ("1.2.3", LevelBump.MAJOR, "2.0.0"),
        ("1.2.3-rc.1", LevelBump.NO_RELEASE, "1.2.3-rc.1"),
        ("1.2.3-rc.1", LevelBump.PRERELEASE_REVISION, "1.2.3-rc.2"),
        ("1.2.3-rc.1", LevelBump.PATCH, "1.2.4-rc.1"),
        ("1.2.3-rc.1", LevelBump.MINOR, "1.3.0-rc.1"),
        ("1.2.3-rc.1", LevelBump.MAJOR, "2.0.0-rc.1"),
    ],
)
def test_version_bump_succeeds(current_version, level, new_version):
    cv = Version.parse(current_version)
    nv = cv.bump(level)
    assert nv == Version.parse(new_version)
    assert cv + level == Version.parse(new_version)


@pytest.mark.parametrize("bad_level", [5, "patch", {"major": True}, [1, 1, 0, 0, 1], 1])
def test_version_bump_typeerror(bad_level):
    with pytest.raises(TypeError):
        Version.parse("1.2.3").bump(bad_level)


def test_version_hashable(a_version):
    _ = {a_version: 4}
    assert True


# NOTE: this might be a really good first candidate for hypothesis
@pytest.mark.parametrize(
    "major, minor, patch, prerelease_revision",
    [tuple(random.choice(range(1, 100)) for _ in range(4)) for _ in range(10)],
)
def test_prerelease_always_less_than_full(major, minor, patch, prerelease_revision):
    full = Version(major, minor, patch)
    pre = Version(major, minor, patch, prerelease_revision=prerelease_revision)
    assert pre < full



"""
Parses commit messages using `scipy tags <scipy-style>`_ of the form::

    <tag>(<scope>): <subject>

    <body>


The elements <tag>, <scope> and <body> are optional. If no tag is present, the
commit will be added to the changelog section "None" and no version increment
will be performed.

While <scope> is supported here it isn't actually part of the scipy style.
If it is missing, parentheses around it are too. The commit should then be
of the form::

    <tag>: <subject>

    <body>

To communicate a breaking change add "BREAKING CHANGE" into the body at the
beginning of a paragraph. Fill this paragraph with information how to migrate
from the broken behavior to the new behavior. It will be added to the
"Breaking" section of the changelog.

Supported Tags::

    (
        API,
        DEP,
        ENH,
        REV,
        BUG,
        MAINT,
        BENCH,
        BLD,
    )
    DEV, DOC, STY, TST, REL, FEAT, TEST

Supported Changelog Sections::

    breaking, feature, fix, Other, None

.. _`scipy-style`: https://docs.scipy.org/doc/scipy/reference/dev/contributor/development_workflow.html#writing-the-commit-message
"""

from __future__ import annotations

import re
from functools import reduce
from itertools import zip_longest
from re import compile as regexp
from textwrap import dedent
from typing import TYPE_CHECKING, Tuple

from git.objects.commit import Commit
from pydantic.dataclasses import dataclass

from semantic_release.commit_parser._base import CommitParser, ParserOptions
from semantic_release.commit_parser.token import (
    ParsedCommit,
    ParsedMessageResult,
    ParseError,
    ParseResult,
)
from semantic_release.commit_parser.util import (
    deep_copy_commit,
    force_str,
    parse_paragraphs,
)
from semantic_release.enums import LevelBump
from semantic_release.errors import InvalidParserOptions
from semantic_release.globals import logger
from semantic_release.helpers import sort_numerically, text_reducer

if TYPE_CHECKING:  # pragma: no cover
    from git.objects.commit import Commit


def _logged_parse_error(commit: Commit, error: str) -> ParseError:
    logger.debug(error)
    return ParseError(commit, error=error)


tag_to_section = {
    "API": "breaking",
    "BENCH": "none",
    "BLD": "fix",
    "BUG": "fix",
    "DEP": "breaking",
    "DEV": "none",
    "DOC": "documentation",
    "ENH": "feature",
    "MAINT": "fix",
    "REV": "other",
    "STY": "none",
    "TST": "none",
    "REL": "none",
    # strictly speaking not part of the standard
    "FEAT": "feature",
    "TEST": "none",
}


@dataclass
class ScipyParserOptions(ParserOptions):
    """
    Options dataclass for ScipyCommitParser

    Scipy-style commit messages follow the same format as Angular-style commit
    just with different tag names.
    """

    major_tags: Tuple[str, ...] = ("API", "DEP")
    """Commit-type prefixes that should result in a major release bump."""

    minor_tags: Tuple[str, ...] = ("ENH", "FEAT")
    """Commit-type prefixes that should result in a minor release bump."""

    patch_tags: Tuple[str, ...] = ("BLD", "BUG", "MAINT")
    """Commit-type prefixes that should result in a patch release bump."""

    other_allowed_tags: Tuple[str, ...] = (
        # "REV",  # Revert commits are NOT Currently Supported
        "DEV",
        "BENCH",
        "DOC",
        "STY",
        "TST",
        "REL",
        "TEST",
    )
    """Commit-type prefixes that are allowed but do not result in a version bump."""

    allowed_tags: Tuple[str, ...] = (
        *major_tags,
        *minor_tags,
        *patch_tags,
        *other_allowed_tags,
    )
    """
    All commit-type prefixes that are allowed.

    These are used to identify a valid commit message. If a commit message does not start with
    one of these prefixes, it will not be considered a valid commit message.
    """

    # TODO: breaking v11, make consistent with AngularParserOptions
    default_level_bump: LevelBump = LevelBump.NO_RELEASE
    """The minimum bump level to apply to valid commit message."""

    parse_squash_commits: bool = True
    """Toggle flag for whether or not to parse squash commits"""

    ignore_merge_commits: bool = True
    """Toggle flag for whether or not to ignore merge commits"""

    @property
    def tag_to_level(self) -> dict[str, LevelBump]:
        """A mapping of commit tags to the level bump they should result in."""
        return self._tag_to_level

    def __post_init__(self) -> None:
        # TODO: breaking v11, remove as the name is now consistent
        self.default_bump_level = self.default_level_bump
        self._tag_to_level: dict[str, LevelBump] = {
            str(tag): level
            for tag, level in [
                # we have to do a type ignore as zip_longest provides a type that is not specific enough
                # for our expected output. Due to the empty second array, we know the first is always longest
                # and that means no values in the first entry of the tuples will ever be a LevelBump. We
                # apply a str() to make mypy happy although it will never happen.
                *zip_longest(self.allowed_tags, (), fillvalue=self.default_bump_level),
                *zip_longest(self.patch_tags, (), fillvalue=LevelBump.PATCH),
                *zip_longest(self.minor_tags, (), fillvalue=LevelBump.MINOR),
                *zip_longest(self.major_tags, (), fillvalue=LevelBump.MAJOR),
            ]
            if "|" not in str(tag)
        }


class ScipyCommitParser(CommitParser[ParseResult, ScipyParserOptions]):
    """Parser for scipy-style commit messages"""

    # TODO: Deprecate in lieu of get_default_options()
    parser_options = ScipyParserOptions

    def __init__(self, options: ScipyParserOptions | None = None) -> None:
        super().__init__(options)

        try:
            commit_type_pattern = regexp(
                r"(?P<type>%s)" % str.join("|", self.options.allowed_tags)
            )
        except re.error as err:
            raise InvalidParserOptions(
                str.join(
                    "\n",
                    [
                        f"Invalid options for {self.__class__.__name__}",
                        "Unable to create regular expression from configured commit-types.",
                        "Please check the configured commit-types and remove or escape any regular expression characters.",
                    ],
                )
            ) from err

        self.commit_prefix = regexp(
            str.join(
                "",
                [
                    f"^{commit_type_pattern.pattern}",
                    r"(?::[\t ]*(?P<scope>[^:\n]+))?",
                    r":[\t ]+",
                ],
            )
        )

        self.commit_msg_pattern = regexp(
            str.join(
                "",
                [
                    self.commit_prefix.pattern,
                    r"(?P<subject>[^\n]+)",
                    r"(?:\n\n(?P<text>.+))?",  # commit body
                ],
            ),
            flags=re.DOTALL,
        )

        # GitHub & Gitea use (#123), GitLab uses (!123), and BitBucket uses (pull request #123)
        self.mr_selector = regexp(
            r"[\t ]+\((?:pull request )?(?P<mr_number>[#!]\d+)\)[\t ]*$"
        )
        self.issue_selector = regexp(
            str.join(
                "",
                [
                    r"^(?:clos(?:e|es|ed|ing)|fix(?:es|ed|ing)?|resolv(?:e|es|ed|ing)|implement(?:s|ed|ing)?):",
                    r"[\t ]+(?P<issue_predicate>.+)[\t ]*$",
                ],
            ),
            flags=re.MULTILINE | re.IGNORECASE,
        )
        self.notice_selector = regexp(r"^NOTICE: (?P<notice>.+)$")
        self.filters = {
            "typo-extra-spaces": (regexp(r"(\S)  +(\S)"), r"\1 \2"),
            "git-header-commit": (
                regexp(r"^[\t ]*commit [0-9a-f]+$\n?", flags=re.MULTILINE),
                "",
            ),
            "git-header-author": (
                regexp(r"^[\t ]*Author: .+$\n?", flags=re.MULTILINE),
                "",
            ),
            "git-header-date": (
                regexp(r"^[\t ]*Date: .+$\n?", flags=re.MULTILINE),
                "",
            ),
            "git-squash-heading": (
                regexp(
                    r"^[\t ]*Squashed commit of the following:.*$\n?",
                    flags=re.MULTILINE,
                ),
                "",
            ),
            "git-squash-commit-prefix": (
                regexp(
                    str.join(
                        "",
                        [
                            r"^(?:[\t ]*[*-][\t ]+|[\t ]+)?",  # bullet points or indentation
                            commit_type_pattern.pattern + r"\b",  # prior to commit type
                        ],
                    ),
                    flags=re.MULTILINE,
                ),
                # move commit type to the start of the line
                r"\1",
            ),
        }

    @staticmethod
    def get_default_options() -> ScipyParserOptions:
        return ScipyParserOptions()

    def commit_body_components_separator(
        self, accumulator: dict[str, list[str]], text: str
    ) -> dict[str, list[str]]:
        if (match := self.notice_selector.match(text)) and (
            notice := match.group("notice")
        ):
            accumulator["notices"].append(notice)
            return accumulator

        if match := self.issue_selector.search(text):
            # if match := self.issue_selector.search(text):
            predicate = regexp(r",? and | *[,;/& ] *").sub(
                ",", match.group("issue_predicate") or ""
            )
            # Almost all issue trackers use a number to reference an issue so
            # we use a simple regexp to validate the existence of a number which helps filter out
            # any non-issue references that don't fit our expected format
            has_number = regexp(r"\d+")
            new_issue_refs: set[str] = set(
                filter(
                    lambda issue_str, validator=has_number: validator.search(issue_str),  # type: ignore[arg-type]
                    predicate.split(","),
                )
            )
            if new_issue_refs:
                accumulator["linked_issues"] = sort_numerically(
                    set(accumulator["linked_issues"]).union(new_issue_refs)
                )
                return accumulator

        # Prevent appending duplicate descriptions
        if text not in accumulator["descriptions"]:
            accumulator["descriptions"].append(text)

        return accumulator

    def parse_message(self, message: str) -> ParsedMessageResult | None:
        if not (parsed := self.commit_msg_pattern.match(message)):
            return None

        parsed_scope = parsed.group("scope") or ""
        parsed_subject = parsed.group("subject")
        parsed_text = parsed.group("text")
        parsed_type = parsed.group("type")

        linked_merge_request = ""
        if mr_match := self.mr_selector.search(parsed_subject):
            linked_merge_request = mr_match.group("mr_number")
            parsed_subject = self.mr_selector.sub("", parsed_subject).strip()

        body_components: dict[str, list[str]] = reduce(
            self.commit_body_components_separator,
            [
                # Insert the subject before the other paragraphs
                parsed_subject,
                *parse_paragraphs(parsed_text or ""),
            ],
            {
                "descriptions": [],
                "notices": [],
                "linked_issues": [],
            },
        )

        level_bump = self.options.tag_to_level.get(
            parsed_type, self.options.default_bump_level
        )

        return ParsedMessageResult(
            bump=level_bump,
            type=parsed_type,
            category=tag_to_section.get(parsed_type, "None"),
            scope=parsed_scope,
            descriptions=tuple(
                body_components["descriptions"]
                if level_bump != LevelBump.MAJOR
                else [parsed_subject]
            ),
            breaking_descriptions=tuple(
                body_components["descriptions"][1:]
                if level_bump == LevelBump.MAJOR
                else []
            ),
            release_notices=tuple(body_components["notices"]),
            linked_issues=tuple(body_components["linked_issues"]),
            linked_merge_request=linked_merge_request,
        )

    @staticmethod
    def is_merge_commit(commit: Commit) -> bool:
        return len(commit.parents) > 1

    def parse_commit(self, commit: Commit) -> ParseResult:
        if not (parsed_msg_result := self.parse_message(force_str(commit.message))):
            return _logged_parse_error(
                commit,
                f"Unable to parse commit message: {commit.message!r}",
            )

        return ParsedCommit.from_parsed_message_result(commit, parsed_msg_result)

    def parse(self, commit: Commit) -> ParseResult | list[ParseResult]:
        """
        Parse a commit message

        If the commit message is a squashed merge commit, it will be split into
        multiple commits, each of which will be parsed separately. Single commits
        will be returned as a list of a single ParseResult.
        """
        if self.options.ignore_merge_commits and self.is_merge_commit(commit):
            return _logged_parse_error(
                commit, "Ignoring merge commit: %s" % commit.hexsha[:8]
            )

        separate_commits: list[Commit] = (
            self.unsquash_commit(commit)
            if self.options.parse_squash_commits
            else [commit]
        )

        # Parse each commit individually if there were more than one
        parsed_commits: list[ParseResult] = list(
            map(self.parse_commit, separate_commits)
        )

        def add_linked_merge_request(
            parsed_result: ParseResult, mr_number: str
        ) -> ParseResult:
            return (
                parsed_result
                if not isinstance(parsed_result, ParsedCommit)
                else ParsedCommit(
                    **{
                        **parsed_result._asdict(),
                        "linked_merge_request": mr_number,
                    }
                )
            )

        # TODO: improve this for other VCS systems other than GitHub & BitBucket
        # Github works as the first commit in a squash merge commit has the PR number
        # appended to the first line of the commit message
        lead_commit = next(iter(parsed_commits))

        if isinstance(lead_commit, ParsedCommit) and lead_commit.linked_merge_request:
            # If the first commit has linked merge requests, assume all commits
            # are part of the same PR and add the linked merge requests to all
            # parsed commits
            parsed_commits = [
                lead_commit,
                *map(
                    lambda parsed_result, mr=lead_commit.linked_merge_request: (  # type: ignore[misc]
                        add_linked_merge_request(parsed_result, mr)
                    ),
                    parsed_commits[1:],
                ),
            ]

        elif isinstance(lead_commit, ParseError) and (
            mr_match := self.mr_selector.search(force_str(lead_commit.message))
        ):
            # Handle BitBucket Squash Merge Commits (see #1085), which have non angular commit
            # format but include the PR number in the commit subject that we want to extract
            linked_merge_request = mr_match.group("mr_number")

            # apply the linked MR to all commits
            parsed_commits = [
                add_linked_merge_request(parsed_result, linked_merge_request)
                for parsed_result in parsed_commits
            ]

        return parsed_commits

    def unsquash_commit(self, commit: Commit) -> list[Commit]:
        # GitHub EXAMPLE:
        # feat(changelog): add autofit_text_width filter to template environment (#1062)
        #
        # This change adds an equivalent style formatter that can apply a text alignment
        # to a maximum width and also maintain an indent over paragraphs of text
        #
        # * docs(changelog-templates): add definition & usage of autofit_text_width template filter
        #
        # * test(changelog-context): add test cases to check autofit_text_width filter use
        #
        # `git merge --squash` EXAMPLE:
        # Squashed commit of the following:
        #
        # commit 63ec09b9e844e616dcaa7bae35a0b66671b59fbb
        # Author: codejedi365 <codejedi365@gmail.com>
        # Date:   Sun Oct 13 12:05:23 2024 -0600
        #
        #     feat(release-config): some commit subject
        #

        # Return a list of artificial commits (each with a single commit message)
        return [
            # create a artificial commit object (copy of original but with modified message)
            Commit(
                **{
                    **deep_copy_commit(commit),
                    "message": commit_msg,
                }
            )
            for commit_msg in self.unsquash_commit_message(force_str(commit.message))
        ] or [commit]

    def unsquash_commit_message(self, message: str) -> list[str]:
        normalized_message = message.replace("\r", "").strip()

        # split by obvious separate commits (applies to manual git squash merges)
        obvious_squashed_commits = self.filters["git-header-commit"][0].split(
            normalized_message
        )

        separate_commit_msgs: list[str] = reduce(
            lambda all_msgs, msgs: all_msgs + msgs,
            map(self._find_squashed_commits_in_str, obvious_squashed_commits),
            [],
        )

        return list(filter(None, separate_commit_msgs))

    def _find_squashed_commits_in_str(self, message: str) -> list[str]:
        separate_commit_msgs: list[str] = []
        current_msg = ""

        for paragraph in filter(None, message.strip().split("\n\n")):
            # Apply filters to normalize the paragraph
            clean_paragraph = reduce(text_reducer, self.filters.values(), paragraph)

            # remove any filtered (and now empty) paragraphs (ie. the git headers)
            if not clean_paragraph.strip():
                continue

            # Check if the paragraph is the start of a new angular commit
            if not self.commit_prefix.search(clean_paragraph):
                if not separate_commit_msgs and not current_msg:
                    # if there are no separate commit messages and no current message
                    # then this is the first commit message
                    current_msg = dedent(clean_paragraph)
                    continue

                # append the paragraph as part of the previous commit message
                if current_msg:
                    current_msg += f"\n\n{dedent(clean_paragraph)}"
                # else: drop the paragraph
                continue

            # Since we found the start of the new commit, store any previous commit
            # message separately and start the new commit message
            if current_msg:
                separate_commit_msgs.append(current_msg)

            current_msg = clean_paragraph

        return [*separate_commit_msgs, current_msg]



from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING
from unittest import mock

import pytest
import shellingham
import tomlkit
from flatdict import FlatDict
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from tests.const import MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures.repos import repo_w_trunk_only_conventional_commits
from tests.util import assert_successful_exit_code, get_func_qual_name

if TYPE_CHECKING:
    from tests.conftest import RunCliFn
    from tests.fixtures.example_project import GetWheelFileFn, UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import BuiltRepoResult


@pytest.mark.skipif(sys.platform == "win32", reason="Unix only")
@pytest.mark.parametrize(
    "shell",
    filter(
        None,
        [
            # because we will actually run the build command in this shell, we must ensure it exists
            "bash"
            if list(
                filter(
                    lambda sh_exe: Path(sh_exe).exists(),
                    ("/bin/bash", "/usr/bin/bash", "/usr/local/bin/bash"),
                )
            )
            else "",
            "zsh"
            if list(
                filter(
                    lambda sh_exe: Path(sh_exe).exists(),
                    ("/bin/zsh", "/usr/bin/zsh", "/usr/local/bin/zsh"),
                )
            )
            else "",
        ],
    )
    or ["sh"],
)
@pytest.mark.parametrize(
    "repo_result, cli_args, next_release_version",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            ["--patch"],
            "0.1.2",
        )
    ],
)
def test_version_runs_build_command(
    repo_result: BuiltRepoResult,
    cli_args: list[str],
    next_release_version: str,
    run_cli: RunCliFn,
    shell: str,
    get_wheel_file: GetWheelFileFn,
    example_pyproject_toml: Path,
    mocked_git_fetch: mock.MagicMock,
    mocked_git_push: mock.MagicMock,
    post_mocker: mock.Mock,
):
    # Setup
    built_wheel_file = get_wheel_file(next_release_version)
    pyproject_config = FlatDict(
        tomlkit.loads(example_pyproject_toml.read_text(encoding="utf-8")),
        delimiter=".",
    )
    build_command = pyproject_config.get("tool.semantic_release.build_command", "")
    patched_os_environment = {
        "CI": "true",
        "PATH": os.getenv("PATH", ""),
        "HOME": "/home/username",
        "VIRTUAL_ENV": "./.venv",
        # Simulate that all CI's are set
        "GITHUB_ACTIONS": "true",
        "GITLAB_CI": "true",
        "GITEA_ACTIONS": "true",
        "BITBUCKET_REPO_FULL_NAME": "python-semantic-release/python-semantic-release.git",
        "PSR_DOCKER_GITHUB_ACTION": "true",
    }

    # Wrap subprocess.run to capture the arguments to the call
    with mock.patch(
        get_func_qual_name(subprocess.run),
        wraps=subprocess.run,
    ) as patched_subprocess_run, mock.patch(
        get_func_qual_name(shellingham.detect_shell), return_value=(shell, shell)
    ):
        # ACT: run & force a new version that will trigger the build command
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *cli_args]
        result = run_cli(cli_cmd[1:], env=patched_os_environment)

        # Evaluate
        assert_successful_exit_code(result, cli_cmd)
        patched_subprocess_run.assert_called_with(
            [shell, "-c", build_command],
            check=True,
            env={
                "NEW_VERSION": next_release_version,  # injected into environment
                "PACKAGE_NAME": "",  # PSR injected environment variable
                "CI": patched_os_environment["CI"],
                "BITBUCKET_CI": "true",  # Converted
                "GITHUB_ACTIONS": patched_os_environment["GITHUB_ACTIONS"],
                "GITEA_ACTIONS": patched_os_environment["GITEA_ACTIONS"],
                "GITLAB_CI": patched_os_environment["GITLAB_CI"],
                "HOME": patched_os_environment["HOME"],
                "PATH": patched_os_environment["PATH"],
                "VIRTUAL_ENV": patched_os_environment["VIRTUAL_ENV"],
                "PSR_DOCKER_GITHUB_ACTION": patched_os_environment[
                    "PSR_DOCKER_GITHUB_ACTION"
                ],
            },
        )

        assert built_wheel_file.exists()
        assert (
            mocked_git_fetch.call_count == 1
        )  # fetch called to check for remote changes
        assert mocked_git_push.call_count == 2
        assert post_mocker.call_count == 1


@pytest.mark.skipif(sys.platform != "win32", reason="Windows only")
@pytest.mark.parametrize("shell", ("powershell", "pwsh", "cmd"))
@pytest.mark.parametrize(
    "repo_result, cli_args, next_release_version",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            ["--patch"],
            "0.1.2",
        )
    ],
)
def test_version_runs_build_command_windows(
    repo_result: BuiltRepoResult,
    cli_args: list[str],
    next_release_version: str,
    run_cli: RunCliFn,
    shell: str,
    get_wheel_file: GetWheelFileFn,
    example_pyproject_toml: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    mocked_git_fetch: mock.MagicMock,
    mocked_git_push: mock.MagicMock,
    post_mocker: mock.Mock,
    clean_os_environment: dict[str, str],
):
    if shell == "cmd":
        build_result_file = get_wheel_file("%NEW_VERSION%")
        update_pyproject_toml(
            "tool.semantic_release.build_command",
            str.join(
                " && ",
                [
                    f"mkdir {build_result_file.parent}",
                    f"type nul > {build_result_file}",
                    f"echo 'Built distribution: {build_result_file}'",
                ],
            ),
        )

    # Setup
    package_name = "my-package"
    update_pyproject_toml("project.name", package_name)
    built_wheel_file = get_wheel_file(next_release_version)
    pyproject_config = FlatDict(
        tomlkit.loads(example_pyproject_toml.read_text(encoding="utf-8")),
        delimiter=".",
    )
    build_command = pyproject_config.get("tool.semantic_release.build_command", "")
    patched_os_environment = {
        **clean_os_environment,
        "CI": "true",
        "VIRTUAL_ENV": "./.venv",
        # Simulate that all CI's are set
        "GITHUB_ACTIONS": "true",
        "GITLAB_CI": "true",
        "GITEA_ACTIONS": "true",
        "BITBUCKET_REPO_FULL_NAME": "python-semantic-release/python-semantic-release.git",
        "PSR_DOCKER_GITHUB_ACTION": "true",
    }

    # Wrap subprocess.run to capture the arguments to the call
    with mock.patch(
        get_func_qual_name(subprocess.run),
        wraps=subprocess.run,
    ) as patched_subprocess_run, mock.patch(
        get_func_qual_name(shellingham.detect_shell), return_value=(shell, shell)
    ):
        # ACT: run & force a new version that will trigger the build command
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, *cli_args]
        result = run_cli(cli_cmd[1:], env=patched_os_environment)

        # Evaluate
        assert_successful_exit_code(result, cli_cmd)
        patched_subprocess_run.assert_called_once_with(
            [shell, "/c" if shell == "cmd" else "-Command", build_command],
            check=True,
            env={
                **clean_os_environment,
                "NEW_VERSION": next_release_version,  # injected into environment
                "PACKAGE_NAME": package_name,  # PSR injected environment variable
                "CI": patched_os_environment["CI"],
                "BITBUCKET_CI": "true",  # Converted
                "GITHUB_ACTIONS": patched_os_environment["GITHUB_ACTIONS"],
                "GITEA_ACTIONS": patched_os_environment["GITEA_ACTIONS"],
                "GITLAB_CI": patched_os_environment["GITLAB_CI"],
                "VIRTUAL_ENV": patched_os_environment["VIRTUAL_ENV"],
                "PSR_DOCKER_GITHUB_ACTION": patched_os_environment[
                    "PSR_DOCKER_GITHUB_ACTION"
                ],
            },
        )

        dist_file_exists = built_wheel_file.exists()
        assert dist_file_exists, f"\n  Expected wheel file to be created at {built_wheel_file}, but it does not exist."
        assert (
            mocked_git_fetch.call_count == 1
        )  # fetch called to check for remote changes
        assert mocked_git_push.call_count == 2
        assert post_mocker.call_count == 1


@pytest.mark.parametrize(
    "repo_result, cli_args, next_release_version",
    [
        (
            lazy_fixture(repo_w_trunk_only_conventional_commits.__name__),
            ["--patch"],
            "0.1.2",
        )
    ],
)
def test_version_runs_build_command_w_user_env(
    repo_result: BuiltRepoResult,
    cli_args: list[str],
    next_release_version: str,
    run_cli: RunCliFn,
    example_pyproject_toml: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    clean_os_environment: dict[str, str],
):
    # Setup
    patched_os_environment = {
        **clean_os_environment,
        "CI": "true",
        "VIRTUAL_ENV": "./.venv",
        # Simulate that all CI's are set
        "GITHUB_ACTIONS": "true",
        "GITLAB_CI": "true",
        "GITEA_ACTIONS": "true",
        "BITBUCKET_REPO_FULL_NAME": "python-semantic-release/python-semantic-release.git",
        "PSR_DOCKER_GITHUB_ACTION": "true",
        # User environment variables (varying passthrough results)
        "MY_CUSTOM_VARIABLE": "custom",
        "IGNORED_VARIABLE": "ignore_me",
        "OVERWRITTEN_VAR": "initial",
        "SET_AS_EMPTY_VAR": "not_empty",
    }
    pyproject_config = FlatDict(
        tomlkit.loads(example_pyproject_toml.read_text(encoding="utf-8")),
        delimiter=".",
    )
    build_command = pyproject_config.get("tool.semantic_release.build_command", "")
    update_pyproject_toml(
        "tool.semantic_release.build_command_env",
        [
            # Includes arbitrary whitespace which will be removed
            " MY_CUSTOM_VARIABLE ",  # detect and pass from environment
            " OVERWRITTEN_VAR = overrided",  # pass hardcoded value which overrides environment
            " SET_AS_EMPTY_VAR = ",  # keep variable initialized but as empty string
            " HARDCODED_VAR=hardcoded ",  # pass hardcoded value that doesn't override anything
            "VAR_W_EQUALS = a-var===condition",  # only splits on 1st equals sign
            "=ignored-invalid-named-var",  # TODO: validation error instead, but currently just ignore
        ],
    )
    package_name = "my-package"
    update_pyproject_toml("project.name", package_name)

    # Mock out subprocess.run
    with mock.patch(
        get_func_qual_name(subprocess.run),
        return_value=subprocess.CompletedProcess(args=(), returncode=0),
    ) as patched_subprocess_run, mock.patch(
        get_func_qual_name(shellingham.detect_shell),
        return_value=("bash", "/usr/bin/bash"),
    ):
        cli_cmd = [
            MAIN_PROG_NAME,
            VERSION_SUBCMD,
            *cli_args,
            "--no-commit",
            "--no-tag",
            "--no-changelog",
            "--no-push",
        ]

        # ACT: run & force a new version that will trigger the build command
        result = run_cli(cli_cmd[1:], env=patched_os_environment)

        # Evaluate
        # [1] Make sure it did not error internally
        assert_successful_exit_code(result, cli_cmd)

        # [2] Make sure the subprocess was called with the correct environment
        patched_subprocess_run.assert_called_once_with(
            ["bash", "-c", build_command],
            check=True,
            env={
                **clean_os_environment,
                "NEW_VERSION": next_release_version,  # injected into environment
                "PACKAGE_NAME": package_name,  # PSR injected environment variable
                "CI": patched_os_environment["CI"],
                "BITBUCKET_CI": "true",  # Converted
                "GITHUB_ACTIONS": patched_os_environment["GITHUB_ACTIONS"],
                "GITEA_ACTIONS": patched_os_environment["GITEA_ACTIONS"],
                "GITLAB_CI": patched_os_environment["GITLAB_CI"],
                "VIRTUAL_ENV": patched_os_environment["VIRTUAL_ENV"],
                "PSR_DOCKER_GITHUB_ACTION": patched_os_environment[
                    "PSR_DOCKER_GITHUB_ACTION"
                ],
                "MY_CUSTOM_VARIABLE": patched_os_environment["MY_CUSTOM_VARIABLE"],
                "OVERWRITTEN_VAR": "overrided",
                "SET_AS_EMPTY_VAR": "",
                "HARDCODED_VAR": "hardcoded",
                # Note that IGNORED_VARIABLE is not here.
                "VAR_W_EQUALS": "a-var===condition",
            },
        )


@pytest.mark.usefixtures(repo_w_trunk_only_conventional_commits.__name__)
def test_version_skips_build_command_with_skip_build(
    run_cli: RunCliFn,
    mocked_git_fetch: mock.MagicMock,
    mocked_git_push: mock.MagicMock,
    post_mocker: mock.Mock,
):
    cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--patch", "--skip-build"]

    with mock.patch(
        get_func_qual_name(subprocess.run),
        return_value=subprocess.CompletedProcess(args=(), returncode=0),
    ) as patched_subprocess_run:
        # Act: force a new version
        result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    patched_subprocess_run.assert_not_called()

    assert mocked_git_fetch.call_count == 1  # fetch called to check for remote changes
    assert mocked_git_push.call_count == 2
    assert post_mocker.call_count == 1



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from requests import HTTPError, Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry  # type: ignore[import]

from semantic_release.globals import logger

if TYPE_CHECKING:  # pragma: no cover
    from semantic_release.hvcs.token_auth import TokenAuth


def build_requests_session(
    raise_for_status: bool = True,
    retry: bool | int | Retry = True,
    auth: TokenAuth | None = None,
) -> Session:
    """
    Create a requests session.

    :param raise_for_status: If True, a hook to invoke raise_for_status be installed
    :param retry: If true, it will use default Retry configuration. if an integer, it
        will use default Retry configuration with given integer as total retry
        count. if Retry instance, it will use this instance.
    :param auth: Optional TokenAuth instance to be used to provide the Authorization
        header to the session

    :return: configured requests Session
    """
    session = Session()
    if raise_for_status:
        session.hooks = {"response": [lambda r, *_, **__: r.raise_for_status()]}

    if retry:
        if isinstance(retry, bool):
            retry = Retry()
        elif isinstance(retry, int):
            retry = Retry(retry)
        elif not isinstance(retry, Retry):
            raise ValueError("retry should be a bool, int or Retry instance.")
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("http://", adapter)
        session.mount("https://", adapter)

    if auth:
        logger.debug("setting up default session authentication")
        session.auth = auth

    return session


_R = TypeVar("_R")


def suppress_http_error_for_codes(
    *codes: int,
) -> Callable[[Callable[..., _R]], Callable[..., _R | None]]:
    """
    For the codes given, return a decorator that will suppress HTTPErrors that are
    raised from responses that came with one of those status codes. The function will
    return False instead of raising the HTTPError
    """

    def _suppress_http_error_for_codes(
        func: Callable[..., _R],
    ) -> Callable[..., _R | None]:
        @wraps(func)
        def _wrapper(*a: Any, **kw: Any) -> _R | None:
            try:
                return func(*a, **kw)
            except HTTPError as err:
                if err.response and err.response.status_code in codes:
                    logger.warning(
                        "%s received response %s: %s",
                        func.__qualname__,
                        err.response.status_code,
                        str(err),
                    )
                return None

        return _wrapper

    return _suppress_http_error_for_codes


suppress_not_found = suppress_http_error_for_codes(404)



from __future__ import annotations

from datetime import timedelta
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest
from git import Commit, Object, Repo

from semantic_release.changelog.release_history import Release, ReleaseHistory
from semantic_release.commit_parser.token import ParsedCommit
from semantic_release.enums import LevelBump
from semantic_release.version.version import Version

if TYPE_CHECKING:
    from git import Actor

    from tests.conftest import GetStableDateNowFn


@pytest.fixture
def artificial_release_history(
    commit_author: Actor,
    stable_now_date: GetStableDateNowFn,
) -> ReleaseHistory:
    current_datetime = stable_now_date()
    first_version = Version.parse("1.0.0")
    second_version = first_version.bump(LevelBump.MINOR)
    fix_commit_subject = "fix a problem"
    fix_commit_type = "fix"
    fix_commit_scope = "cli"

    fix_commit = Commit(
        Repo("."),
        Object.NULL_HEX_SHA[:20].encode("utf-8"),
        message=f"{fix_commit_type}({fix_commit_scope}): {fix_commit_subject}",
    )

    fix_commit_parsed = ParsedCommit(
        bump=LevelBump.PATCH,
        type="fix",
        scope=fix_commit_scope,
        descriptions=[fix_commit_subject],
        breaking_descriptions=[],
        commit=fix_commit,
    )

    fix_commit_2_subject = "alphabetically first to solve a non-scoped problem"
    fix_commit_2_type = "fix"
    fix_commit_2_scope = ""

    fix_commit_2 = Commit(
        Repo("."),
        Object.NULL_HEX_SHA[:20].encode("utf-8"),
        message=f"{fix_commit_2_type}: {fix_commit_2_subject}",
    )

    fix_commit_2_parsed = ParsedCommit(
        bump=LevelBump.PATCH,
        type="fix",
        scope=fix_commit_2_scope,
        descriptions=[fix_commit_2_subject],
        breaking_descriptions=[],
        commit=fix_commit_2,
    )

    fix_commit_3_subject = "alphabetically first to solve a scoped problem"
    fix_commit_3_type = "fix"
    fix_commit_3_scope = "cli"

    fix_commit_3 = Commit(
        Repo("."),
        Object.NULL_HEX_SHA[:20].encode("utf-8"),
        message=f"{fix_commit_3_type}({fix_commit_3_scope}): {fix_commit_3_subject}",
    )

    fix_commit_3_parsed = ParsedCommit(
        bump=LevelBump.PATCH,
        type="fix",
        scope=fix_commit_3_scope,
        descriptions=[fix_commit_3_subject],
        breaking_descriptions=[],
        commit=fix_commit_3,
    )

    feat_commit_subject = "add a new feature"
    feat_commit_type = "feat"
    feat_commit_scope = "cli"

    feat_commit = Commit(
        Repo("."),
        Object.NULL_HEX_SHA[:20].encode("utf-8"),
        message=f"{feat_commit_type}({feat_commit_scope}): {feat_commit_subject}",
    )

    feat_commit_parsed = ParsedCommit(
        bump=LevelBump.MINOR,
        type="feature",
        scope=feat_commit_scope,
        descriptions=[feat_commit_subject],
        breaking_descriptions=[],
        commit=feat_commit,
    )

    return ReleaseHistory(
        unreleased={"feature": [feat_commit_parsed]},
        released={
            second_version: Release(
                tagger=commit_author,
                committer=commit_author,
                tagged_date=current_datetime,
                elements={
                    # Purposefully inserted out of order, should be dictsorted in templates
                    "fix": [
                        # Purposefully inserted out of alphabetical order, should be sorted in templates
                        fix_commit_parsed,
                        fix_commit_2_parsed,  # has no scope
                        fix_commit_3_parsed,  # has same scope as 1
                    ],
                    "feature": [feat_commit_parsed],
                },
                version=second_version,
            ),
            first_version: Release(
                tagger=commit_author,
                committer=commit_author,
                tagged_date=current_datetime - timedelta(minutes=1),
                elements={"feature": [feat_commit_parsed]},
                version=first_version,
            ),
        },
    )


@pytest.fixture
def release_history_w_brk_change(
    artificial_release_history: ReleaseHistory,
    stable_now_date: GetStableDateNowFn,
) -> ReleaseHistory:
    current_datetime = stable_now_date()
    latest_version = next(iter(artificial_release_history.released.keys()))
    next_version = latest_version.bump(LevelBump.MAJOR)
    brk_commit_subject = "fix a problem"
    brk_commit_type = "fix"
    brk_commit_scope = "cli"
    brk_change_msg = "this is a breaking change"

    brk_commit = Commit(
        Repo("."),
        Object.NULL_BIN_SHA,
        message=str.join(
            "\n\n",
            [
                f"{brk_commit_type}({brk_commit_scope}): {brk_commit_subject}",
                f"BREAKING CHANGE: {brk_change_msg}",
            ],
        ),
    )

    brk_commit_parsed = ParsedCommit(
        bump=LevelBump.MAJOR,
        type=brk_commit_type,
        scope=brk_commit_scope,
        descriptions=[brk_commit_subject],
        breaking_descriptions=[brk_change_msg],
        commit=brk_commit,
    )

    return ReleaseHistory(
        unreleased={},
        released={
            next_version: Release(
                tagger=artificial_release_history.released[latest_version]["tagger"],
                committer=artificial_release_history.released[latest_version][
                    "committer"
                ],
                tagged_date=current_datetime,
                elements={"Bug Fixes": [brk_commit_parsed]},
                version=next_version,
            ),
            **artificial_release_history.released,
        },
    )


@pytest.fixture
def release_history_w_multiple_brk_changes(
    release_history_w_brk_change: ReleaseHistory,
    stable_now_date: GetStableDateNowFn,
) -> ReleaseHistory:
    current_datetime = stable_now_date()
    latest_version = next(iter(release_history_w_brk_change.released.keys()))
    brk_commit_subject = "adding a revolutionary feature"
    brk_commit_type = "feat"
    brk_change_msg = "The feature changes everything in a breaking way"

    brk_commit = Commit(
        Repo("."),
        Object.NULL_BIN_SHA,
        message=str.join(
            "\n\n",
            [
                f"{brk_commit_type}: {brk_commit_subject}",
                f"BREAKING CHANGE: {brk_change_msg}",
            ],
        ),
    )

    brk_commit_parsed = ParsedCommit(
        bump=LevelBump.MAJOR,
        type=brk_commit_type,
        scope="",  # No scope in this commit
        descriptions=[brk_commit_subject],
        breaking_descriptions=[brk_change_msg],
        commit=brk_commit,
    )

    return ReleaseHistory(
        unreleased={},
        released={
            **release_history_w_brk_change.released,
            # Replaces and inserts a new commit of different type with breaking changes
            latest_version: Release(
                tagger=release_history_w_brk_change.released[latest_version]["tagger"],
                committer=release_history_w_brk_change.released[latest_version][
                    "committer"
                ],
                tagged_date=current_datetime,
                elements={
                    **release_history_w_brk_change.released[latest_version]["elements"],
                    "Features": [brk_commit_parsed],
                },
                version=latest_version,
            ),
        },
    )


@pytest.fixture
def single_release_history(
    artificial_release_history: ReleaseHistory,
) -> ReleaseHistory:
    version = list(artificial_release_history.released.keys())[-1]
    return ReleaseHistory(
        unreleased={},
        released={
            version: artificial_release_history.released[version],
        },
    )


@pytest.fixture
def release_history_w_a_notice(
    artificial_release_history: ReleaseHistory,
    stable_now_date: GetStableDateNowFn,
) -> ReleaseHistory:
    current_datetime = stable_now_date()
    latest_version = next(iter(artificial_release_history.released.keys()))
    next_version = latest_version.bump(LevelBump.PATCH)
    notice_commit_subject = "deprecate a type"
    notice_commit_type = "refactor"
    notice_commit_scope = "cli"
    release_notice = dedent(
        """\
        This is a multline release notice that is made
        up of two lines.
        """
    )

    notice_commit = Commit(
        Repo("."),
        Object.NULL_BIN_SHA,
        message=str.join(
            "\n\n",
            [
                f"{notice_commit_type}({notice_commit_scope}): {notice_commit_subject}",
                f"NOTICE: {release_notice}",
            ],
        ),
    )

    notice_commit_parsed = ParsedCommit(
        bump=LevelBump.NO_RELEASE,
        type=notice_commit_type,
        scope=notice_commit_scope,
        descriptions=[notice_commit_subject],
        breaking_descriptions=[],
        release_notices=(release_notice.replace("\n", " ").strip(),),
        commit=notice_commit,
    )

    return ReleaseHistory(
        unreleased={},
        released={
            next_version: Release(
                tagger=artificial_release_history.released[latest_version]["tagger"],
                committer=artificial_release_history.released[latest_version][
                    "committer"
                ],
                tagged_date=current_datetime,
                elements={"Refactoring": [notice_commit_parsed]},
                version=next_version,
            ),
            **artificial_release_history.released,
        },
    )


@pytest.fixture
def release_history_w_notice_n_brk_change(
    artificial_release_history: ReleaseHistory,
    release_history_w_a_notice: ReleaseHistory,
    stable_now_date: GetStableDateNowFn,
) -> ReleaseHistory:
    current_datetime = stable_now_date()
    latest_version = next(iter(artificial_release_history.released.keys()))
    next_version = latest_version.bump(LevelBump.MAJOR)
    brk_commit_subject = "fix a problem"
    brk_commit_type = "fix"
    brk_commit_scope = "cli"
    brk_change_msg = "this is a breaking change"

    brk_commit = Commit(
        Repo("."),
        Object.NULL_BIN_SHA,
        message=str.join(
            "\n\n",
            [
                f"{brk_commit_type}({brk_commit_scope}): {brk_commit_subject}",
                f"BREAKING CHANGE: {brk_change_msg}",
            ],
        ),
    )

    brk_commit_parsed = ParsedCommit(
        bump=LevelBump.MAJOR,
        type=brk_commit_type,
        scope=brk_commit_scope,
        descriptions=[brk_commit_subject],
        breaking_descriptions=[brk_change_msg],
        commit=brk_commit,
    )

    last_notice_release = next(iter(release_history_w_a_notice.released.keys()))

    return ReleaseHistory(
        unreleased={},
        released={
            next_version: Release(
                tagger=artificial_release_history.released[latest_version]["tagger"],
                committer=artificial_release_history.released[latest_version][
                    "committer"
                ],
                tagged_date=current_datetime,
                elements={
                    "Bug Fixes": [brk_commit_parsed],
                    **release_history_w_a_notice.released[last_notice_release][
                        "elements"
                    ],
                },
                version=next_version,
            ),
            **artificial_release_history.released,
        },
    )


@pytest.fixture
def release_history_w_multiple_notices(
    release_history_w_a_notice: ReleaseHistory,
    stable_now_date: GetStableDateNowFn,
) -> ReleaseHistory:
    current_datetime = stable_now_date()
    latest_version = next(iter(release_history_w_a_notice.released.keys()))

    notice_commit_subject = "add a configurable feature"
    notice_commit_type = "feat"
    notice_commit_scope = "cli-config"
    release_notice = dedent(
        """\
        This is a multline release notice that is its own
        paragraph to detail the configurable feature.
        """
    )

    notice_commit = Commit(
        Repo("."),
        Object.NULL_BIN_SHA,
        message=str.join(
            "\n\n",
            [
                f"{notice_commit_type}({notice_commit_scope}): {notice_commit_subject}",
                f"NOTICE: {release_notice}",
            ],
        ),
    )

    notice_commit_parsed = ParsedCommit(
        bump=LevelBump.MINOR,
        type=notice_commit_type,
        scope=notice_commit_scope,
        descriptions=[notice_commit_subject],
        breaking_descriptions=[],
        release_notices=(release_notice.replace("\n", " ").strip(),),
        commit=notice_commit,
    )

    return ReleaseHistory(
        unreleased={},
        released={
            **release_history_w_a_notice.released,
            # Replaces and inserts a new commit of different type with breaking changes
            latest_version: Release(
                tagger=release_history_w_a_notice.released[latest_version]["tagger"],
                committer=release_history_w_a_notice.released[latest_version][
                    "committer"
                ],
                tagged_date=current_datetime,
                elements={
                    "Features": [notice_commit_parsed],
                    **release_history_w_a_notice.released[latest_version]["elements"],
                },
                version=latest_version,
            ),
        },
    )
